/* mtmconf.c	Initialisation and configuration of MTM-SUN transputer boards
 *
 * Copyright (C) Parsytec Gmbh, 1990
 *
 * Author:	Matthias Clauss
 *
 * Date:	2.7.1990	Created
 *
 * Version:	1.1
 *
 * Updates:
 *	Vers	date	author	description
 *	1.1	900730	HD	usage of a configuration file implemented
 *
 * Description:
 *	This program initialises the MTM-SUN board.
 *	First a program written in Occam is downloaded to the
 *	NCU.
 *	Second some arrays are built describing the configuration
 *	defined in a configuration file (The configuration file must
 *	be defined at compile time).
 *	Third these arrays are sent to the NCU.
 *
 *	See also file mtmsun.h
 */


/*
	CONFIGURATION_FILE contains a description of the connections
	we wish to establish. Syntax:

	One or more lines like

		connect ( ITEM1 , ITEM2 )

	where ITEM1 and ITEM2 are connectable links. You should only
	use the link macros declared in mtmsun.h . C-style comments
	and preprocessor statements are also allowed.

	You can define CONFIGURATION_FILE in your makefile. Default
	is "mtmsun.ncu".

	For every new configuration you have to call make.
*/

#ifndef CONFIGURATION_FILE
#define CONFIGURATION_FILE	"mtmsun.ncu"
#endif

#include "../link.h"

#include "mtmsun.h"
#include <stdio.h>
#include <signal.h>

#define	BOOT_FILE_1	"/usr/etc/transp/mtm1ncu.cde"	/* for 1 ncu  */
#define	BOOT_FILE_2	"/usr/etc/transp/mtm2ncu.cde"	/* for 2 ncus */


#define	BUFF_SIZE	(4 * 1024)
#define	BOOT_TIMEOUT	10		/* in 1/10 seconds */
#define	WRITE_TIMEOUT	30
#define	READ_TIMEOUT	30
#define	CNTL_TIMEOUT	30

/* exit codes */
#define	EINFOLINK	1	/* */
#define	ELOPEN		2	/* failed to open link */
#define	ENOCONFL	3	/* */
#define EPARAM          4       /* illegal parameter in argv */
#define ECODELOAD	5	/* error while loading code file */
#define EOPENLINK	2
#define ERESETLINK	6
#define ECONNECT	7	/* illegal connection was tried */

#define JUSTFINISH	255	/* exit code for read configuration */

/* external functions */
extern char *lerror ();		/* decodes error message */
extern int   OpenLink ();	/* open link */
extern int   ResetLink ();	/* reset transputer */
extern int   WriteLink ();	/* send data to transputer */
extern int   ReadLink  ();	/* receive data from transputer */

/* local functions */
void usage ();
int configure ();
int code_load ();
int DoNcu ();
int change_vme_timeout ();
int read_configuration ();
void report_success();
void report_failure();
void _connect ();
#define connect(c1, c2) _connect(c1, c2);

/* local variables */
int two_ncus		= FALSE;
int verbose		= FALSE;
int just_read_it	= FALSE;
int free_even_comm_link = channel_id(COMM_0);
int free_odd_comm_link  = channel_id(COMM_1);
char *progname;

int connect_errors = 0;

int main (argc, argv, env)
int argc;
char *argv[], *env[];
{
	char *link_name = NULL;		/* link adaptor name */
	int link_id;			/* used to identify the link */
	int lerr;			/* local error code */

	progname = argv[0];
	while (--argc > 0) {
		if (**++argv == '-') {
			switch (argv[0][1]) {
				case 'v': verbose = TRUE;
				          break;
				case 'r': just_read_it = TRUE;
				          verbose = TRUE;
				          break;
				default:  usage();
				          report_failure();
				          exit(EPARAM);
			}
		} else {
			if (link_name == NULL) {
				link_name = *argv;
			} else {
				usage();
				report_failure();
				exit(EPARAM);
			}
		}
	}
	if (link_name == NULL) {
		usage();
		report_failure();
		exit(EPARAM);
	}

	lerr = configure (link_name, just_read_it);
	if (lerr) {
		if (lerr == JUSTFINISH)
			report_success();
		else
			report_failure();
		exit(lerr);
	}

	/* 'mtmconf' is intended to be called by 'ulti'.
	 * In order to confirm 'ulti' that the configuration has
	 * succeeded, we send a SIGUSR1 signal.
	 * Using the exit value is not appropriate, because
	 * for the future a MTM-SUN ncm will run in and endless
	 * loop.
	 * So here I am going to send SIGUSR1 to my parent
	 * and than fall asleep.
	 */
	report_success();
	pause ();
	return (0);
}



/*
 *	usage
 */

void usage ()
{
	printf ("\nusage: %s [-v | -r] link-name\n", progname);
	printf ("configure Parsytec\'s MTM-SUN transputer board\n");
	printf ("-v : verbose\n-r : read configuration\n");
	fflush (stdout);
}



/*	configure
 *
 *	configure the specified link, which must be the
 *	config_in-link. This will be done by opening the link
 *	and booting the T222 with the occcam executable for the
 *	correct number of ncu's available on that MTM-SUN.
 *	After that the desired topology is switched by sending
 *	data packets according to a protocoll.
 *	
 */

/* the following variables are declared
 * global, in order to initialise them.
 * For future: should be made local.
 *
 *	ncu command structures to reset and configure
 */

static struct to_ncu reset0 = {
	RESET_A,	NCU0,
	{STAY32},
	{STAY32},
	DO_RESET, DO_RESET, DO_RESET
};

static struct to_ncu reset1 = {
	RESET_A,	NCU1,
	{STAY32},
	{STAY32},
	DO_RESET, DO_RESET, DO_RESET
};

static struct to_ncu config0 = {
	CONNECT_A,	NCU0,
	{STAY32},
	{STAY32},
	NO_RESET, NO_RESET, NO_RESET
};

static struct to_ncu config1 = {
	CONNECT_A,	NCU1,
	{STAY32},
	{STAY32},
	NO_RESET, NO_RESET, NO_RESET
};



int configure (link_name, just_read_it)
char *link_name;
int just_read_it;
{
	struct ext_link_info *x_link = NULL;
	int no_links;			/* number of links */
	int link_id;			/* used to identify the link */
	int lerr;			/* local error code */
	int config_type = CONF_NULL;	/* InfoLink() .config_type */

	/* open the specified link and check wether is it
	 * a MTM-SUN1 or MTM-SUN2 config-link.
	 * Rely on LinkId is the index in 'ext_link_info' array.
	 */
	if ((no_links = InfoLink (&x_link)) < 0) {
		/* failed to get info about links */
		fprintf (stderr,
			 "\n%s: Failed to get info about links, because %s\n",
			 progname, lerror (no_links));
		return (EINFOLINK);
	}
	lerr = OpenLink (link_name);
	if ( lerr < 0 ) {
		/* failed to open link */
		fprintf (stderr,
			 "\n%s: Failed to open link \"%s\", because %s\n",
			 progname, link_name, lerror (lerr));
		return (ELOPEN);
	}
	/* store returned link-id */
	link_id = lerr;

	config_type = x_link[link_id].config_type;
	if ( (config_type != CONF_MTMSUN1) &&
	     (config_type != CONF_MTMSUN2) ) {
	     /* config type neither MTM-SUN1 nor MTM-SUN2 */
	     fprintf(stderr,
		     "\n%s: Link \"%s\" is no MTM-SUN config-link (type %d)\n",
		     progname, link_name, config_type);
	     return (ENOCONFL);
	}
	/* now it is sure, that link-name is a config-link for MTM-SUN 1/2
	 * that means config_type == CONF_MTMSUN1 || CONF_MTMSUN2.
	 * Reset ncu transputer (T222) and download corresponding code_file.
	 */
	two_ncus = (config_type == CONF_MTMSUN2);
	lerr = code_load (link_id, two_ncus? BOOT_FILE_2: BOOT_FILE_1);
	if (lerr != OK) {
		/* failed to load the code file */
		fprintf (stderr,
			 "\n%s: Failed to load the code file\n",
			 progname);
		return (ECODELOAD);
	}

	if (just_read_it) {
		printf("\n%s: read configuration of %s\n", 
		       progname, link_name);
		lerr = DoNcu (link_id, &config0, "Read onboard    ncu");
		if (two_ncus) {
			/* read also additional ncu */
			lerr = DoNcu (link_id, &config1, 
				      "read additional ncu");
		}
		return(JUSTFINISH);
	} else {
		/* send reset request to ncu transputer */
		lerr = DoNcu (link_id, &reset0, "Reset onboard    ncu");
		if (two_ncus) {
			/* reset also additional ncu */
			lerr = DoNcu (link_id, &reset1, 
				      "Reset additional ncu");
		}

		if ((lerr = read_configuration()) != 0) {
			fprintf(stderr, 
				"\n%s: %d error(s) occured (aborted)!\n", 
				progname, lerr);
			return(ECONNECT);
		}

		lerr = DoNcu (link_id, &config0, "Configure onboard    ncu");
		if (two_ncus) {
			/* configure also additional ncu */
			lerr = DoNcu (link_id, &config1, 
				      "Configure additional ncu");
		}
/*
		do NOT close that link. During the pause() call keep the 
		link locked
		CloseLink (link_id);
*/
	}
	return (0);
}


/*
	_connect() is called for every connection we wish to establish.
	features:

	- error detection
	- automatic use of communication links between on-board-ncu and
	  add-on-ncu (if you have one)
*/

static void _connect(c1, c2)
int c1, c2;
{
	struct to_ncu *ncu_command = NULL;
	int odd;
	int help;
	int comm_link;
	static int additional_ncu_error_printed = 0;

	/* check for even-odd-conection */
	if ((odd = odd_ncu(c1)) != odd_ncu(c2)) {
		fprintf(stderr,
			"\n%s: connection between even and odd ncu not", 
			progname);
		fprintf(stderr, "\npossible: %4.4X %4.4X\n", c1, c2);
		connect_errors++;
		return;
	}

	/* usage of communication links not allowed */
	if (communication_link(c1) || communication_link(c2)) {
		fprintf(stderr, 
			"\n%s: usage of communication links (COMM_x)", 
			progname);
		fprintf(stderr, "\nnot allowed\n");
		connect_errors++;
		return;
	}

	/* some external connectors can only be connected to the */
	/* additional ncu, iff you have one. The logical conversion */
	/* is made here. */
	if (two_ncus) {
		if (is_ext_special(c1))
			c1 = convert_ext(c1);
		if (is_ext_special(c2))
			c2 = convert_ext(c2);
	}

	/* if the links use the same ncu */
	if (ncu_board(c1) == ncu_board(c2)) {
		/* ncu available ? */
		if (!two_ncus && ncu_board(c1) == NCU_BOARD_1) {
			if (!additional_ncu_error_printed) {
				fprintf(stderr, 
				 "\nError: additional ncu is not available\n");
				additional_ncu_error_printed = 1;
				connect_errors++;
			}
			return;
		}
		/* make the entry in the right structure */
		ncu_command = (ncu_board(c1) == NCU_BOARD_0)? 
			&config0 : &config1;
		if (odd)
			ncu_command->odd[channel_id(c1)] = channel_id(c2);
		else
			ncu_command->even[channel_id(c1)] = channel_id(c2);
		/* we have done it! */
		return;
	}

	/* if the links use not the same ncu */
	if (!two_ncus) {
		if (!additional_ncu_error_printed) {
			fprintf(stderr, 
				"\n%s: additional ncu is not available\n",
				progname);
			additional_ncu_error_printed = 1;
			connect_errors++;
		}
		return;
	}

	/* make the work easier */
	if (ncu_board(c1) == NCU_BOARD_1) {
		help = c1;
		c1 = c2;
		c2 = help;
	}

	/* Now c1 uses ncu board 0 and c2 uses ncu board 1. */
	/* Get a communication link and make the connection */
	/* entries in the correct structures */
	if (odd) {
		comm_link = free_odd_comm_link++;
		if (comm_link > channel_id(COMM_11)) {
			fprintf(stderr, 
				"\n%s: No more communication links available",
				progname);
			fprintf(stderr, "\nfor odd ncu\n");
			connect_errors++;
			return;
		}
		config0.odd[channel_id(c1)] = comm_link;
		config1.odd[channel_id(c2)] = comm_link;
	} else {
		comm_link = free_even_comm_link++;
		if (comm_link > channel_id(COMM_10)) {
			fprintf(stderr, 
				"\n%s: No more communication links available",
				progname);
			fprintf(stderr, "\nfor even ncu\n");
			connect_errors++;
			return;
		}
		config0.even[channel_id(c1)] = comm_link;
		config1.even[channel_id(c2)] = comm_link;
	}
}


/*
	read_configuration() makes all the connections the
	user has described in CONFIGURATION_FILE. For every
	connection the function _connect() (see above) is
	called.

	The return value is the number of errors.

*/

int read_configuration()
{
	connect_errors = 0;
#include CONFIGURATION_FILE
	return(connect_errors);
}



/*	code_load
 *
 *	Resets the transputer and downloads the specified
 *	transputer executable file.
 *	Returns OK on success. Error diagnostics are printed
 *	directly using stderr.
 */

int code_load (link_id, boot_file)
int link_id;
char *boot_file;
{
	FILE *fp;			/* boot-file */
	int lerr;			/* local error code */
	int cnt;
	int total_send;			/* bytes send so far */
	unsigned char sdata, rdata;	/* send/receive-data for transputer */
	char fbuffer[BUFF_SIZE];	/* boot-file buffer */

	/* reset transputer */
	lerr = ResetLink (link_id);
	if ( lerr < 0 ) {
		/* failed to reset link */
		fprintf (stderr,
			"\n%s: Failed to reset link %d, because %s\n",
			progname, link_id, lerror (lerr));
		return (!OK);
	}

	/* download transputer code-file */
	fp = fopen (boot_file, "r");
	if ( fp == NULL ) {
		/* failed to open boot file */
		fprintf (stderr,
			"\n%s: Failed to open boot-file \"%s\", because ",
			progname, boot_file);
		perror ("");
		return (!OK);
	}
	/* boot file opened successfully */
	total_send = 0;
	while ((cnt = fread (fbuffer, 1, sizeof (fbuffer), fp)) > 0) {
		/* try to send fbuffer to transputer */
		lerr = WriteLink (link_id, fbuffer, cnt, BOOT_TIMEOUT);
		if ( lerr != cnt ) {
			/* failed to send complete data */
			if ( lerr >= 0 ) {
				/* timeout during send */
				fprintf (stderr,
				   "\n%s: Only %d from %d bytes send\n",
				   progname, lerr, cnt);
			} else {
				/* failed to send */
				fprintf (stderr,
				"\n%s: Failed to send at 0x%X, because %s\n",
				progname, total_send, lerror (lerr));
			}
			return (!OK);
		}
		total_send += cnt;
	}
	/* check for fread error */
	if ( feof (fp) ) {
		/* end of file ? That's correct. */
		fclose (fp);
	} else {
		/* error during read */
		fprintf (stderr,
			"\n%s: Read error from boot-file \"%s\",",
			progname, boot_file);
		perror (" because ");
		return (!OK);
	}
	/* everything is paletti */
	return (OK);
}

/*	DoNcu
 *	Handle configuration with ncu's.
 * protocoll to ncu is :
 *	word	reset.a
 *	word	<NCU #>
 *	byte	dummy[32]
 *	byte	dummy[32]
 *	byte	1		/* reset even link switch
 *	byte	1		/* reset even link switch
 *	byte	1		/* reset hardware
 * receive as answer :
 *	word	<tag == return.b>
 *	word	<NCu #>
 *	byte	dummy[32]
 *	byte	dummy[32]
 *	long	<config-time in us>
 *	word	error_code	/* 0 <-> OK
 *
 *	Note:
 *	=====
 *	Unfortunatly it is NOT possible, to use 'sizeof (struct to_ncu)'
 *	or 'sizeof (struct from_ncu)', due to word alignement of
 *	structures. So I have choosen the usage of explicit size-values
 *	defined as symbolic values.
 */

int DoNcu (link_id, data, message)
int   link_id;
struct to_ncu *data;
char *message;
{
	int	lerr;			/*local return value */
	int	i;			/* debugging */
	byte	*dt;			/* debugging */
	struct from_ncu from_ncu;	/* ncu reply */

	/* send configuration vector */
	lerr = WriteLink (link_id, data, SIZE_TO_NCU, WRITE_TIMEOUT);
	if ( lerr <= 0 ) {
		/* error/timeout during send */
		fprintf(stderr,
			"\n%s: %s Failed to send,  because %s\n",
			progname, message, 
			(lerr == 0) ? "timeout" : lerror (lerr));
		return (!OK);
	}

	/* receive answer from NCU */
	lerr = ReadLink (link_id, &from_ncu, SIZE_FROM_NCU, READ_TIMEOUT);
	if ( lerr <= 0 ) {
		/* error/timeout during receive */
		fprintf(stderr,
			"\n%s: %s Failed to receive, because %s\n",
			progname, message, 
			(lerr == 0) ? "timeout" : lerror (lerr));
		return (!OK);
	}

	/* check answer half hearted */
	if ( from_ncu.err_code != 0 ) {
		/* first error handling print error code */
		fprintf(stderr,
			"\n%s: %s ncu return error 0x%x\n",
			progname, message, from_ncu.err_code);
		return (!OK);
	}

	if (verbose) {
		printf ("\n%s: From Ncu %04x :", message, from_ncu.ncu);
		printf ("\nEven NCU: ");
		for (i=0, dt=from_ncu.even; i<32; i++, dt++) {
			if ((i & 0x0F) == 0)
				printf("\n");
			if ( *dt & 0x80 )
				printf ("%02d ", *dt-0x80);
			else
				printf ("-- ");
		}
		printf ("\nOdd  NCU: ");
		for (i=0, dt=from_ncu.odd; i<32; i++, dt++) {
			if ((i & 0x0F) == 0)
				printf("\n");
			if ( *dt & 0x80 )
				printf ("%02d ", *dt-0x80);
			else
				printf ("-- ");
		}
		printf ("\n");
	}

	/* everything is perfectly fine. Close link */
	return (OK);
}


/*
	'mtmconf' is intended to be called by 'ulti'.
	In order to confirm 'ulti' that the configuration has
	succeeded, we send a SIGUSR1 signal, otherwise a 
	SIGUSR2 signal.

	Using the exit value is not appropriate, because
	for the future a MTM-SUN ncm will run in and endless
	loop.
*/

void report_success()
{
	if ( kill (getppid (), SIGUSR1) == -1 ) {
	    /* failed to send signal */
	    fprintf (stderr, 
		 "\n%s: Failed to send signal SIGUSR1 to parent process\n");
	         progname, perror ("because, ");
        }
}


void report_failure()
{
	if ( kill (getppid (), SIGUSR2) == -1 ) {
	    /* failed to send signal */
	    fprintf (stderr, 
		 "\n%s: Failed to send signal SIGUSR2 to parent process\n");
		 progname, perror ("because, ");
	}
}

