/*
 * mtmsun.h		MTM-SUN transputer board declarations
 *
 * Copyright (C) Parsytec GmbH, 1990.
 *
 * Author:      Matthias Clauss
 *
 * Date:        5.3.1990        Created
 *
 * Version:     1.02
 *
 * Updates:     Vers    date    author  description
 *		1.01	22.6.90	mc	descripton of TRANSP_ enhanced
 *		1.02	24.7.90 hd	link macros
 *
 */

/*	The MTM-SUN board may be equipped with an additional
 *	"add-on" ncu. This will slightly change some physical
 *	link numbers.
 */

/*
 *	Mapping of physical links to units
 *	Due to even/odd link-switches, every even link may be
 *	accessed by the even NCU only, for odd links only by
 *	the odd NCU.
 *	Besides that, additional functionality is given by
 *	upgrading the board with an additional NCU board.
 */

/*	Onboard transputer and first 4 link-adaptor to VME bus
 *
 *	Note: Only the first 4 of total 8 link-adaptors may
 *	      be used with the onboard-ncu.
 *	      The additional ncu is required for the other
 *	      4 link-adaptors.
 *
 *	See also: board-name in 'trans_conf' file.
 */


#define _CHANNEL_MASK		0x1F
#define _EVEN			0x0000
#define _ODD			0x0100
#define _NCU0			0x0200
#define _NCU1			0x0400
#define _EXT_SPECIAL		0x0800
#define _NCU_MASK		_ODD
#define _NCU_BOARD_MASK 	(_NCU0 | _NCU1)
#define NCU_BOARD_0		0
#define NCU_BOARD_1		1



/*
	channel_id(link_id) returns the channel number of a
	specified link
*/
#define channel_id(c1)		((c1) & _CHANNEL_MASK)



/*
	odd_ncu(link_id) returns !=0 if the specified link
	uses the odd ncu
*/
#define odd_ncu(c1)	       	(((c1) & _NCU_MASK) == _ODD)



/*
	communication_link(link_id) returns !=0 if the specified
	link is a communication link.
*/
#define communication_link(c1)						\
	(((c1) & _NCU_BOARD_MASK) == _NCU_BOARD_MASK)



/*
	ncu_board(link_id) returns NCU_BOARD_0 if the specified
	link can be connected to the on-board-ncu.

	different_ncu_boards(link_1, link_2) returns !=0 if the
	specified links can only be connected via a communication
	link.
*/
#define ncu_board(c1)							\
	((((c1) & _NCU_BOARD_MASK) == _NCU0)? NCU_BOARD_0: NCU_BOARD_1)
#define different_ncu_boards(c1, c2)					\
	(((c1) & (c2) & _NCU_BOARD_MASK) == 0)



/*
	Some external connectors are switched to the add-on-ncu,
	iff you have one. The following macros handle the conversion.

	is_ext_special(link_id) returns !=0 if the spec. link is one
	of these link connectors.

	convert_ext(link_id) converts the link_id of an external connector
	for the usage with the add-on-ncu. Should only be used, when
	is_ext_special(link_id) succeeds and you have an add-on-ncu.
*/
#define is_ext_special(c1)	(((c1) & _EXT_SPECIAL) != 0)
#define convert_ext(c1)		((((c1) & ~_NCU0) | _NCU1) - 4)



/*
	To configure your MTM-Sun board you have to write a description
	file (normally named mtmsun.ncu). It contains a description of
	the connections you wish to establish. Syntax:

	One or more lines like

		connect ( ITEM1 , ITEM2 )

	where ITEM1 and ITEM2 are connectable links. You should only
	use the link macros declared below. C-style comments are also
	allowed.
*/

/*		link_id		channel	Even/Odd ncu */

#define		TRANSP_0	(0	| _EVEN	| _NCU0)
#define		TRANSP_1	(0	| _ODD	| _NCU0)
#define		TRANSP_2	(1	| _EVEN	| _NCU0)
#define		TRANSP_3	(1	| _ODD	| _NCU0)

#define		LINK_0		(18	| _EVEN	| _NCU0)
#define		LINK_1		(18	| _ODD	| _NCU0)
#define		LINK_2		(19	| _EVEN	| _NCU0)
#define		LINK_3		(19	| _ODD	| _NCU0)


/*	DBT slots
 *	DBT_slot_link
 */

#define		DBT_0_0		(2	| _EVEN	| _NCU0)
#define		DBT_0_1		(2	| _ODD	| _NCU0)
#define		DBT_0_2		(3	| _EVEN	| _NCU0)
#define		DBT_0_3		(3	| _ODD	| _NCU0)

#define		DBT_1_0		(4	| _EVEN	| _NCU0)
#define		DBT_1_1		(4	| _ODD	| _NCU0)
#define		DBT_1_2		(5	| _EVEN	| _NCU0)
#define		DBT_1_3		(5	| _ODD	| _NCU0)

#define		DBT_2_0		(6	| _EVEN	| _NCU0)
#define		DBT_2_1		(6	| _ODD	| _NCU0)
#define		DBT_2_2		(7	| _EVEN	| _NCU0)
#define		DBT_2_3		(7	| _ODD	| _NCU0)

#define		DBT_3_0		(8	| _EVEN	| _NCU0)
#define		DBT_3_1		(8	| _ODD	| _NCU0)
#define		DBT_3_2		(9	| _EVEN	| _NCU0)
#define		DBT_3_3		(9	| _ODD	| _NCU0)

#define		DBT_4_0		(10	| _EVEN	| _NCU0)
#define		DBT_4_1		(10	| _ODD	| _NCU0)
#define		DBT_4_2		(11	| _EVEN	| _NCU0)
#define		DBT_4_3		(11	| _ODD	| _NCU0)

#define		DBT_5_0		(12	| _EVEN	| _NCU0)
#define		DBT_5_1		(12	| _ODD	| _NCU0)
#define		DBT_5_2		(13	| _EVEN	| _NCU0)
#define		DBT_5_3		(13	| _ODD	| _NCU0)

#define		DBT_6_0		(14	| _EVEN	| _NCU0)
#define		DBT_6_1		(14	| _ODD	| _NCU0)
#define		DBT_6_2		(15	| _EVEN	| _NCU0)
#define		DBT_6_3		(15	| _ODD	| _NCU0)

#define		DBT_7_0		(16	| _EVEN	| _NCU0)
#define		DBT_7_1		(16	| _ODD	| _NCU0)
#define		DBT_7_2		(17	| _EVEN	| _NCU0)
#define		DBT_7_3		(17	| _ODD	| _NCU0)


/*	MultiCluster slots
 *	Notice: MultiCluster slots overlap some DBT slots!
 *	MC_slot_link
 */

#define		MC_0_0		(10	| _EVEN	| _NCU0)
#define		MC_0_1		(10	| _ODD	| _NCU0)
#define		MC_0_2		(11	| _EVEN	| _NCU0)
#define		MC_0_3		(11	| _ODD	| _NCU0)
#define		MC_0_4		(12	| _EVEN	| _NCU0)
#define		MC_0_5		(12	| _ODD	| _NCU0)
#define		MC_0_6		(13	| _EVEN	| _NCU0)
#define		MC_0_7		(13	| _ODD	| _NCU0)

#define		MC_1_0		(14	| _EVEN	| _NCU0)
#define		MC_1_1		(14	| _ODD	| _NCU0)
#define		MC_1_2		(15	| _EVEN	| _NCU0)
#define		MC_1_3		(15	| _ODD	| _NCU0)
#define		MC_1_4		(16	| _EVEN	| _NCU0)
#define		MC_1_5		(16	| _ODD	| _NCU0)
#define		MC_1_6		(17	| _EVEN	| _NCU0)
#define		MC_1_7		(17	| _ODD	| _NCU0)


/*	External Lemosa (round) connectors
 */

#define		LEMO_0		(20	| _EVEN	| _NCU0)
#define		LEMO_1		(20	| _ODD	| _NCU0)

/*	External flat cable connectors
 *	Ext_connector_link
 */

#define		EXT_0_0		(22	| _EVEN	| _NCU0)
#define		EXT_0_1		(22	| _ODD	| _NCU0)
#define		EXT_0_2		(23	| _EVEN	| _NCU0)
#define		EXT_0_3		(23	| _ODD	| _NCU0)

#define		EXT_1_4		(24	| _EVEN	| _NCU0)
#define		EXT_1_5		(24	| _ODD	| _NCU0)
#define		EXT_1_6		(25	| _EVEN	| _NCU0)
#define		EXT_1_7		(25	| _ODD	| _NCU0)

#define		EXT_2_8		(26	| _EVEN	| _NCU0	| _EXT_SPECIAL)
#define		EXT_2_9		(26	| _ODD	| _NCU0	| _EXT_SPECIAL)
#define		EXT_2_10	(27	| _EVEN	| _NCU0	| _EXT_SPECIAL)
#define		EXT_2_11	(27	| _ODD	| _NCU0	| _EXT_SPECIAL)

#define		EXT_3_12	(28	| _EVEN	| _NCU0	| _EXT_SPECIAL)
#define		EXT_3_13	(28	| _ODD	| _NCU0	| _EXT_SPECIAL)
#define		EXT_3_14	(29	| _EVEN	| _NCU0	| _EXT_SPECIAL)
#define		EXT_3_15	(29	| _ODD	| _NCU0	| _EXT_SPECIAL)


/*	Internal communication link between NCU's
 *	Note: Overlapping of communication links and Ext connectors
 *	Valid if both NCU's are present.
 *	One more note: You haven't to deal with the communication
 *	links. This is done for you by the program in mtmconf.c
 */

#define		COMM_0		(26	| _EVEN	| _NCU0	| _NCU1)
#define		COMM_1		(26	| _ODD	| _NCU0	| _NCU1)
#define		COMM_2		(27	| _EVEN	| _NCU0	| _NCU1)
#define		COMM_3		(27	| _ODD	| _NCU0	| _NCU1)
#define		COMM_4		(28	| _EVEN	| _NCU0	| _NCU1)
#define		COMM_5		(28	| _ODD	| _NCU0	| _NCU1)
#define		COMM_6		(29	| _EVEN	| _NCU0	| _NCU1)
#define		COMM_7		(29	| _ODD	| _NCU0	| _NCU1)
#define		COMM_8		(30	| _EVEN	| _NCU0	| _NCU1)
#define		COMM_9		(30	| _ODD	| _NCU0	| _NCU1)
#define		COMM_10		(31	| _EVEN	| _NCU0	| _NCU1)
#define		COMM_11		(31	| _ODD	| _NCU0	| _NCU1)


/*	The following connections require the second NCU
 *	to be installed.
 */

#define		LINK_4		(18	| _EVEN	| _NCU1)
#define		LINK_5		(18	| _ODD	| _NCU1)
#define		LINK_6		(19	| _EVEN	| _NCU1)
#define		LINK_7		(19	| _ODD	| _NCU1)


/*	DBT slots (second ncu is required)
 *	DBT_slot_link
 */

#define		DBT_8_0		(2	| _EVEN	| _NCU1)
#define		DBT_8_1		(2	| _ODD	| _NCU1)
#define		DBT_8_2		(3	| _EVEN	| _NCU1)
#define		DBT_8_3		(3	| _ODD	| _NCU1)

#define		DBT_9_0		(4	| _EVEN	| _NCU1)
#define		DBT_9_1		(4	| _ODD	| _NCU1)
#define		DBT_9_2		(5	| _EVEN	| _NCU1)
#define		DBT_9_3		(5	| _ODD	| _NCU1)

#define		DBT_10_0	(6	| _EVEN	| _NCU1)
#define		DBT_10_1	(6	| _ODD	| _NCU1)
#define		DBT_10_2	(7	| _EVEN	| _NCU1)
#define		DBT_10_3	(7	| _ODD	| _NCU1)

#define		DBT_11_0	(8	| _EVEN	| _NCU1)
#define		DBT_11_1	(8	| _ODD	| _NCU1)
#define		DBT_11_2	(9	| _EVEN	| _NCU1)
#define		DBT_11_3	(9	| _ODD	| _NCU1)

#define		DBT_12_0	(10	| _EVEN	| _NCU1)
#define		DBT_12_1	(10	| _ODD	| _NCU1)
#define		DBT_12_2	(11	| _EVEN	| _NCU1)
#define		DBT_12_3	(11	| _ODD	| _NCU1)

#define		DBT_13_0	(12	| _EVEN	| _NCU1)
#define		DBT_13_1	(12	| _ODD	| _NCU1)
#define		DBT_13_2	(13	| _EVEN	| _NCU1)
#define		DBT_13_3	(13	| _ODD	| _NCU1)

#define		DBT_14_0	(14	| _EVEN	| _NCU1)
#define		DBT_14_1	(14	| _ODD	| _NCU1)
#define		DBT_14_2	(15	| _EVEN	| _NCU1)
#define		DBT_14_3	(15	| _ODD	| _NCU1)



/*	Protocol description to basic software running
 *	on the ncu's.
 *	Note: Be sure to have this software alive
 *
 *	byte	8  bits
 *	word	16 bits (lsb first)
 *	long	32 bits (lsb first)
 *	bool	8  bits (0 FALSE, 1 TRUE)
 *
 *	The protocoll consists of a cycle :
 *	Send fixed length data packet
 *		receive fixed sized data packet
 *
 *	The words "to" and "from" are seen from host
 *	standpoint of view.
 *
 *	The physical connections can be defined in the
 *	byte arrays 'even' and 'odd'. Example: To connect
 *	link adaptor 2 (channel number 19) with node 
 *	transputer link 2 (channel number 1) you have to
 *	set the 19th element of the even-array to the
 *	value 1. Since all lines are bidirectional you can
 *	also define the 1st element of the even-array to the
 *	value 19. You surely can use the #defines above,
 *	but you must know their meanings to use them.
 *
 *	To leave a connection unchanged just insert STAY.
 *	Several connections can be left unchanged with
 *	STAY8, STAY16 and STAY32.
 *
 ****************************************************************
 *								*
 *	It is not possible to connect an even transputer	*
 *	link with an odd one, so take care !!!			*
 *								*
 ****************************************************************
 *
 *	An easier way to define the connections is to
 *	use a configuration file (see mtmconf.c).
 */

typedef	unsigned char	bool;
typedef	unsigned short	word;

struct to_ncu {
	word	to_tag;		/* cmd spec */
	word	ncu;		/* ncu # 0:onboard, 1 add-on */
	byte	even[32];	/* phys conections even ncu */
	byte	odd[32];	/* phys conections odd ncu */
	bool	res_even;	/* to_tag == reset. reset even ncu */
	bool	res_odd;	/* to_tag == reset. reset odd ncu */
	bool	res_hardw;	/* to_tag == reset. reset hardware */
};

struct from_ncu {
	word	from_tag;	/* cmd spec */
	word	ncu;		/* ncu # 0:onboard, 1 add-on */
	byte	even[32];	/* phys conections even ncu */
	byte	odd[32];	/* phys conections odd ncu */
	long	conf_time;	/* time needed for configuration in us */
	word	err_code;	/* bit coded error. 0 <-> OK */
};

#define	SIZE_TO_NCU	71
#define	SIZE_FROM_NCU	74

/* ncu numbers */

#define	NCU0		((word)0x0000)
#define	NCU1		((word)0x0100)

/*	to_tag values */

#define	CONNECT_A	((word)(0x2100))
#define	DISCONNECT_A	((word)(0x2200))
#define	RESET_A		((word)(0x2300))
#define	RETURN_A	((word)(0x2400))
#define	STOP_REQ_A	((word)(0x2500))

/*	from_tag values */

#define	RETURN_B	((word)(0x2400))
#define	STOP_ACK_B	((word)(0x2500))

/*	the value in even[]/odd[] (struct 'to_ncu') for a connection,
 *	which should not be changed.
 */

#define	STAY		((byte)0x70)

#define	STAY4		STAY, STAY, STAY, STAY 
#define	STAY8		STAY, STAY, STAY, STAY, STAY, STAY, STAY, STAY 
#define	STAY16		STAY8, STAY8
#define	STAY32		STAY16, STAY16

/* bool reset values */

#define	DO_RESET		((byte)1)
#define	NO_RESET		((byte)0)


/* return values */

#define	OK	1
#define	NERR	2


/*
 *	define a nice value for even/odd ncu.
 */

#define EVEN	0
#define ODD	1

