/* -*-C-*- */
/*
	acctest.c

	demo program for using vmeacc.h

	900808 HD for parsytec


	This little demo program shows how to use
	the VMEbus access macros in vmeacc.h. It
	assumes to be running on a bbk-v2 and tries
	to access the dual ported memory of another
	bbk-v2 with the VMEbus base address 0x90000000.
*/

#include <stdio.h>
#define BBK_V2
#include "vmeacc.h"

/*
	example: accessing dual ported memory of 
	another bbk-v2 on 0x90000000. Remember:
	You can not access the first 2048 or 4096
	Bytes, since the on-chip RAM of the transputer
	on the other board is mapped in here.
*/
#define DESTINATION_BASE	0x90001000


/*
	raw 8 bit dump. Simply shows the contents of
	16 bytes starting at first_address.
*/
void raw_dump(int first_address)
{
	int i;
	char k;
	printf("\nraw dump: ");
	for (i=0; i<16; i+=sizeof(k)) {
		vme_byte_rd_8ba(k, first_address+i);
		printf("%2.2X  ", k);
	}
}


/*
	show 8 bit access
*/
void byte_test(int first_address)
{
	int i;
	char k;
	printf("\n8 bit access :");

	printf("\nwriting : ");
	for (i=0; i<16; i++) {
		k = 0x03*i;
		printf("%2.2X  ", k&0xFF);
		vme_byte_wr_8ba(k, first_address+i);
	}

	printf("\nreading : ");
	for (i=0; i<16; i++) {
		vme_byte_rd_8ba(k, first_address+i);
		printf("%2.2X  ", k&0xFF);
	}

	raw_dump(first_address);
	putchar('\n');
}


/*
	show 8 bit array access: writing with
	8 bit access mode, reading with 8 and
	32 bit access mode.
*/
void byte_ar_test(int first_address)
{
	int i;
	char k[64];
	printf("\n8 bit array access : look out for your format adaption");

	printf("\nwriting : ");
	for (i=0; i<64; i+=4) {
		k[i] = 0x03*(i>>2);
		printf("%2.2X  ", k[i]&0xFF);
	}
	printf("\t8ba");
	vme_byte_wr_ar_8ba(k, sizeof(k), first_address);

	printf("\nreading : ");
	vme_byte_rd_ar_8ba(k, sizeof(k), first_address);
	for (i=0; i<64; i+=4) {
		printf("%2.2X  ", k[i]&0xFF);
	}
	printf("\t8ba");

	printf("\nreading : ");
	vme_byte_rd_ar_32ba(k, 16, first_address);
	for (i=0; i<16; i++) {
		printf("%2.2X  ", k[i]&0xFF);
	}
	printf("\t32ba !");

	raw_dump(first_address);
	putchar('\n');
}


/*
	show 16 bit access
*/
void int16_test(int first_address)
{
	int i;
	short k;
	printf("\n16 bit access :");

	printf("\nwriting : ");
	for (i=0; i<16; i+=2) {
		k = 0x1234*(i+1);
		printf("%4.4X    ", k&0xFFFF);
		vme_int16_wr_16ba(k, first_address+i);
	}

	printf("\nreading : ");
	for (i=0; i<16; i+=2) {
		vme_int16_rd_16ba(k, first_address+i);
		printf("%4.4X    ", k&0xFFFF);
	}

	raw_dump(first_address);
	putchar('\n');
}


/*
	show 16 bit array access: writing with
	16 bit access mode, reading with 16 and
	32 bit access mode.
*/
void int16_ar_test(int first_address)
{
	int i;
	short k[32];
	printf("\n16 bit array access : look out for your format adaption");

	printf("\nwriting : ");
	for (i=2; i<32; i+=4) {
		k[i] = 0x1234*(i+1);
		printf("%4.4X    ", k[i]&0xFFFF);
	}
	printf("\t16ba");
	vme_int16_wr_ar_16ba(k, sizeof(k), first_address);

	printf("\nreading : ");
	vme_int16_rd_ar_16ba(k, sizeof(k), first_address);
	for (i=2; i<32; i+=4) {
		printf("%4.4X    ", k[i]&0xFFFF);
	}
	printf("\t16ba");

	printf("\nreading : ");
	vme_int16_rd_ar_32ba(k, 8*sizeof(short), first_address);
	for (i=0; i<8; i++) {
		printf("%4.4X    ", k[i]&0xFFFF);
	}
	printf("\t32ba !");

	raw_dump(first_address);
	putchar('\n');
}


/*
	show 32 bit access
*/
void int_test(int first_address)
{
	int i, k;
	printf("\n32 bit access :");

	printf("\nwriting : ");
	for (i=0; i<16; i+=4) {
		k = 0x12345678*(i+1);
		printf("%8.8X        ", k);
		vme_int32_wr_32ba(k, first_address+i);
	}

	printf("\nreading : ");
	for (i=0; i<16; i+=4) {
		vme_int32_rd_32ba(k, first_address+i);
		printf("%8.8X        ", k);
	}

	raw_dump(first_address);
	putchar('\n');
}


/*
	show 32 bit array access: writing and reading
	with 32 bit access mode.
*/
void int_ar_test(int first_address)
{
	int i;
	int k[4];
	printf("\n32 bit array access :");

	printf("\nwriting : ");
	for (i=0; i<4; i++) {
		k[i] = 0x12345678*(i+1);
		printf("%8.8X        ", k[i]);
	}
	vme_int32_wr_ar_32ba(k, sizeof(k), first_address);

	printf("\nreading : ");
	vme_int32_rd_ar_32ba(k, sizeof(k), first_address);
	for (i=0; i<4; i++) {
		printf("%8.8X        ", k[i]);
	}

	raw_dump(first_address);
	putchar('\n');
}


/*
	looping round the memory
*/
void memory_loop(int first_address)
{
	int i = first_address-1;

	printf("\nmemory_loop : (it will take a minute)");
	fflush(stdout);
	do {
		i++;
		*vme_byte_8ba(i)=0x5A;
	} while (*vme_byte_8ba(i) == 0x5A);
	printf("\rmemory_loop : %8.8X bytes successfully tested\n", 
	       i-first_address);
}


int main(void)
{
	/* extended non-privileged data access */
	vme_address_modifier(0x09);

	/* very important: set base address */
	vme_address_register(DESTINATION_BASE);

	/* demo functions, see above */
	byte_test(DESTINATION_BASE);
	int16_test(DESTINATION_BASE);
	int_test(DESTINATION_BASE);
	byte_ar_test(DESTINATION_BASE);
	int16_ar_test(DESTINATION_BASE);
	int_ar_test(DESTINATION_BASE);

	memory_loop(DESTINATION_BASE);
	return(0);
}
