/*{{{ COMMENT Standard Header*/
/******************************************************************
COPYRIGHT (C) INMOS LIMITED - 1990, 1991

    Source file name : clilib.c
              Author : S.R.Christian
   Date first issued : 12-Jun-90


This module contains the core support routines of the c language cli

Date        Modification                                    Initials
----        ------------                                    --------
12-Jun-90   Created module from existing BCPL harness       SRC
01-Oct-91   Added generic functions for formatted I/O       SRC
07-Oct-91   Added conditional compilation for UNHOSTED
            operation. This removes all references to file
            functions and deletes commands which use files  SRC
******************************************************************/
/*}}}*/

/* use reduced library for UNHOSTED operation and declare arbitrary 'stdin' */ 
#ifdef UNHOSTED
#include <stdiored.h>
#define stdin (FILE*)1
#else
#include <stdio.h>
#endif

/* for gcc only */
#ifndef FILENAME_MAX
#define FILENAME_MAX 64
#endif

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include "clilib.h"

/*{{{static variables*/
static symtab_t    **symtab_p;     /* symbol table pointer */
static long        max_symbol_cnt; /* maximum number of symbols */
static long        symbol_cnt;     /* number of symbols declared */
static short       macro_depth;    /* macro invocation depth */
static short       number_base;    /* default input/output base for ints */

static FILE        *cur_stream_p;  /* current file ip stream */
static macro_t     *cur_macro_p;   /* current macro position */
static char        *prompt_p;      /* input prompt */
static boolean     cur_echo;       /* status of input echo */

/* declare an output formating buffer for use by print calls */
static char        format_buf[256];/* limited output length */

/* decare a file descriptor for logging command input */
static FILE*        logfiled;
/* declare a standard command statement parameter separator set */
static char delim_set[] = " ,\\";
/*}}}*/

/*{{{print*/
/* generic output routed through user declared function */
extern int io_write(char* buffer); 

void print(const char*  format, ...)
/* indirect print method using variable arg format to invoke formating */
{
   va_list list;

   va_start(list, format);
   vsprintf(format_buf, format, list);
   va_end(list);
   io_write(format_buf);
}
/*}}}*/
/*{{{is_delim*/
boolean is_delim(char character, char *delim_p)
/* tests the character for equality with a set of delimiter characters */
{
   short delim_count = 0;

   while ((delim_p[delim_count] != '\0') && 
          (character != delim_p[delim_count]))
      delim_count++;
   return ((character == delim_p[delim_count]) ||
           (character == nl_char) ||
           (character == cr_char) ||
           (character == '\0'));
}
/*}}}*/
/*{{{is_matched*/
boolean is_matched(char *tested_p, char *definition_p, short minlen)
/* tests strings for equality, but not in an exact way. Trailing space is OK
   Comparison will succeed if the tested string matches the definition string
   but is shorter (i.e is an abbreviation). It will not match if the tested
   string is longer or less than minlen chars are present in the tested 
   string. Comparison is also case insensitive. */
{
  boolean match = true;
  short cnt = 0;
 
  while (((tested_p[cnt] == definition_p[cnt]) ||
          ((tested_p[cnt] & 0xdf) == (definition_p[cnt] & 0xdf)) ) &&
         (tested_p[cnt] != '\0')                                   &&
         (definition_p[cnt] != '\0'))
    cnt++;
  
  /* if we found the end of the tested string before we found a mis-match
     then strings are matched. If the definition string is shorter than
     minumum length requirements, then match can succeed. */
  if ((tested_p[cnt] != '\0') ||
      ((cnt < minlen) && (definition_p[cnt] != '\0')))        
    match = false;
  
  return(match);
}
/*}}}*/
/*{{{is_control*/
boolean is_control(char *token_p, short *construct_p)
/* tests a token against a set of control redirection primitives 
   and returns an identifier for the construct found */
{
  boolean found = true;

  if (is_matched(token_p, "DEFINE", 2))
    *construct_p = define_const;
  else if (is_matched(token_p, "IF", 2))
    *construct_p = if_const;
  else if (is_matched(token_p, "ELSE", 2))
    *construct_p = else_const;
  else if (is_matched(token_p, "WHILE", 2))
    *construct_p = while_const;
  else if (is_matched(token_p, "FOR", 2))
    *construct_p = for_const;
  else
    found = false;
  return(found);
}
/*}}}*/
/*{{{conv_int*/
long conv_int(char *token_p, short default_base)
/* returns, if possible, an integer value. The default base
   is used for conversion in the absence of other information.
   If a '#' character preceeds the number hex base is assumed 
   If a '$' is used as a prefix, then a binary representation is assumed
   If an 'o' or 'O' caharcter is used then octal is assumed.
   Any sign makes the number a decimal representation */
{
  long    value;
  short   base, cnt;
  boolean negative;
  static  char  *conv = "0123456789ABCDEF";

  negative = false;
  if (token_p[0] == '#') {
    token_p++;
    base = 16;
  }
  else if (token_p[0] == '$'){
    token_p++;
    base = 2;
  }
  else if (token_p[0] == 'o'){
    token_p++;
    base = 8;
  }
  else if (token_p[0] == 'O'){
    token_p++;
    base = 8;
  }
  else if (token_p[0] == '-'){
    token_p++;
    negative = true;
    base = 10;
  }
  else if (token_p[0] == '+'){
    token_p++;
    base = 10;
  }
  else base = default_base;
  
  /* convert by comparison to ordered string array of numeric characters */

  value = 0;
  cnt = 0;
  while((token_p[cnt] != '\0') && (value > bad_integer_val)) {
    short i=0;
    while((conv[i] != (char)toupper(token_p[cnt])) && (i < base))
      i++;
    if ((i >= base) || (value >= (long) 0x7fffffff))
      value = bad_integer_val;
    else
      value = (value * base) + i;
    cnt++;
  }
  if (negative && (value > bad_integer_val)) value = -value;
  
  return(value);
}
/*}}}*/
/*{{{conv_flt*/
boolean conv_flt(char *token_p, double *value_p)
/* returns, if possible, an floating point value. Scanf seems not to
   detect any errors in its input format, and so we have to do a check
   on number validity prior to calling it. This is rather complicated
   to try and reject float expressions, but validate pure float numbers */
{
  short   cnt = 0;
  boolean error = false;
  boolean seen_dp = false;
 
  while((token_p[cnt] != '\0') && !error) {
    /* check for silly characters */
    if((token_p[cnt] != '-') &&
       (token_p[cnt] != '+') &&
       (token_p[cnt] != '.') &&
       (token_p[cnt] != 'E') &&
       (token_p[cnt] != 'e') &&
       ((token_p[cnt] > '9') || (token_p[cnt] < '0'))) error = true;
    /* check for more than one decimal point */
    if (token_p[cnt] == '.') {
      if (seen_dp) error = true;
      else seen_dp = true;
    }
    /* check for sign after decimal point not associated with exponent */
    if (((token_p[cnt] == '+') || (token_p[cnt] == '-')) && seen_dp){
      if ((token_p[cnt-1] != 'E') && (token_p[cnt-1] != 'e')) error = true;
    }
    /* check for sign before a decimal point but not at start of token */
    if (((token_p[cnt] == '+') || (token_p[cnt] == '-')) && !seen_dp){
      if (cnt > 0) error = true;
    }
    cnt++; 
  } 
  if (error || (sscanf(token_p, "%lg", value_p) == 0)) 
    *value_p = bad_integer_val;

  return(error);
}
/*}}}*/
/*{{{look_for*/
symtab_t *look_for(char *token_p, short type)
/* looks for a symbol type/name within the table. We look from
   most recent to oldest symbol to accomodate scope.
   We also scan for the shortest defined symbol that matches
   the input token - this resolves order dependent declaration problems*/
{
  short     cnt;
  boolean   found = false;
  symtab_t  *symbol_p;
  symtab_t  *shortest_p;
  short     short_len;
  
  short_len = max_tok_len;
  shortest_p = (symtab_t*)NULL;

  /* point to last symbol in the table */
  if (symbol_cnt != 0) {
    cnt = 1;
    while(cnt <= symbol_cnt) {
      symbol_p = symtab_p[symbol_cnt - cnt];
      /* look for a name match of at least two characters and 
         shortest matching definition of search type */
      found = ((is_matched(token_p, symbol_p->name_p, 2) &&
               ((symbol_p->type & type) > 0)));
      if (found && (symbol_p->name_len < short_len)) {
        shortest_p = symbol_p;
        short_len = symbol_p->name_len;
        /* print("shortest %s, len=%d\n", 
            symbol_p->name_p, symbol_p->name_len); */
      }
      cnt++;
    }
  }
  return(shortest_p);
}
/*}}}*/
/*{{{tag_current_line*/
void tag_current_line(parse_t *pars_p, char *message_p)
/* displays current tokeniser string and tags the position of last token.
   An optional message is displayed */
{
  char tag[120];
  short i;

  print("%s\n",pars_p->line_p);
  for(i=0; i<pars_p->par_pos; i++)
    tag[i] = space_char;  
  tag[pars_p->par_pos] = '^';
  tag[pars_p->par_sta] = '^';
  tag[(pars_p->par_pos) + 1] = '\0';
  print("%s\n%s\n", tag, message_p);
}
/*}}}*/
/*{{{pars_debug*/
void pars_debug(parse_t *pars_p) 
/* included for tokeniser testing */
{
  tag_current_line(pars_p, "debug");
  print("Tok = \"%s\", delim = %x, toklen = %d \n",
         pars_p->token, 
         pars_p->tok_del, 
         pars_p->tok_len);
}
/*}}}*/
/*{{{init_pars*/
void init_pars(parse_t *pars_p, char *new_line_p)
/* start up a new environment */
{
  pars_p->line_p = new_line_p;
  pars_p->par_pos = 0;
  pars_p->par_sta = 0;
  pars_p->tok_del = '\0';
  pars_p->token[0] = '\0';
  pars_p->tok_len = 0;
}
/*}}}*/
/*{{{cp_pars*/
void cp_pars(parse_t *dest_p, parse_t *source_p)
/* copies parsing status */
{
  *dest_p = *source_p;
}
/*}}}*/
/*{{{get_tok*/
short get_tok(parse_t *pars_p, char *delim_p)
/* implements a tokeniser with a 'soft' approach to end of line conditions -  
   repeated calls will eventually leave parse position at the null terminating
   the input string. */
{
  short par_sta = pars_p->par_pos;
  short par_pos = par_sta;
  short tok_len = 0;
  short quotes = 0;

  /* check that we are not already at the end of a line due to
     a previous call (or a null input) - if so return a null token */

  if (pars_p->line_p[par_pos] == '\0') {
    pars_p->token[0] = '\0';
    pars_p->tok_del = '\0';
  }
  else {
    /* attempt to find start of a token, noting special case of first call,
       incrementing past last delimiter and checking for end of line
       on the way */
    if (par_pos != 0) par_pos++;  
    while (((pars_p->line_p[par_pos] == space_char) ||
            (pars_p->line_p[par_pos] == tab_char)) &&
           (pars_p->line_p[par_pos] != '\0'))
      par_pos++;
    
    /* if we find a delimiter before anything else, return a null token */
    if (is_delim(pars_p->line_p[par_pos], delim_p)) {
      pars_p->token[0] = '\0';
      pars_p->tok_del = pars_p->line_p[par_pos];
    }
    else {
      /* copy token from line into token string
         until next delimiter found. Note that delimiters found within
         pairs of double quotes will not be considered significant. Quotes
         can be embedded within strings using '\' however*/
      while((!is_delim(pars_p->line_p[par_pos], delim_p) || (quotes > 0)) &&
            (tok_len < max_tok_len) && (pars_p->line_p[par_pos] != '\0')) {
        pars_p->token[tok_len] = pars_p->line_p[par_pos++];
        if ((pars_p->token[tok_len] =='"') &&
            (pars_p->token[tok_len-1] !='\\')) {
          if (quotes > 0)
            quotes--;
          else
            quotes++;
        }
        tok_len++;
      }  
      /* if we ran out of token space before copy ended, move up to delimiter */
      while (!is_delim(pars_p->line_p[par_pos], delim_p))
        par_pos++;
    
      /* tidy up the rest of the data */
      pars_p->tok_del = pars_p->line_p[par_pos];
      pars_p->token[tok_len] = '\0';
    }
  }
  pars_p->par_pos = par_pos;
  pars_p->par_sta = par_sta;
  pars_p->tok_len = tok_len;
  
  /* pars_debug(pars_p); */    /* testing only */
  
  return(tok_len);
}
/*}}}*/
/*{{{init_sym_table*/
void init_sym_table(long elements)
/* sets up a simple symbol table as an array of elements, ordered by
   declaration, of different types. symbol_cnt will index the next free slot */
{
  symbol_cnt = 0;
  max_symbol_cnt = elements; /* based on mimimum size required */
  symtab_p = (symtab_t**) malloc((size_t)(sizeof(symtab_t*) * max_symbol_cnt));
}
/*}}}*/
/*{{{purge_symbols*/
void purge_symbols(short level)
/* deletes all symbol entries in table created above a given nest level.
   if the last entry is deleted, the table store is deallocated. */
{
  long      cnt;
  symtab_t  *symbol_p;
  boolean   exit = false;

  /* point to last symbol in the table */
  
  cnt = 1;
  while(!exit && (cnt <= symbol_cnt)) {
    symbol_p = symtab_p[symbol_cnt - cnt];
    /* free up symbol contents before freeing symbol storeage */
    if (!(exit = (symbol_p->depth <= level))) {
      /*{{{delete macro line contents*/
      if (symbol_p->type == mac_symbol) {
        macro_t  *l_p,*nl_p;
        
        l_p = symbol_p->value.mac_val;
        while (l_p != (macro_t*)NULL) {
          nl_p = l_p->line_p;
          free((char*)l_p);
          l_p = nl_p;
        }
      }
      /*}}}*/
      free((char*)symbol_p);
      cnt++;
    }
  }
  symbol_cnt = symbol_cnt - (cnt-1);
  if (symbol_cnt == 0) free((char*) symtab_p);
}
/*}}}*/
/*{{{insert_symbol*/
symtab_t *insert_symbol(char *token_p, short type)
/* adds symbol structure to table and checks for name clashes, scope
   clashes, invalid identifiers and lack of table space. A valid name
   must contain alphanumerics only, plus '_', with 
   the first character alphabetic. */
{
  symtab_t   *symbol_p;
  symtab_t   *oldsym_p;
  short      cnt;
  boolean    valid;

  /* check that symbol table has a spare slot and issue a warning if close */
  symbol_p = (symtab_t*) NULL;
  if (symbol_cnt >= (max_symbol_cnt - 1))
    print("Cannot insert \"%s\" in symbol table - table is full!\n",token_p);
  else {
    valid = true;
    cnt = 0;
    while (token_p[cnt] != '\0') {
      if (((token_p[cnt]  < 'A') || (token_p[cnt] > 'Z')) &&
          ((token_p[cnt]  < 'a') || (token_p[cnt] > 'z')) &&
          ((token_p[cnt]  < '0') || (token_p[cnt] > '9') || (cnt == 0)) &&
          ((token_p[cnt] != '_') || (cnt == 0))) 
        valid = false;
      cnt++;
    }
    if (!valid)
      print("Cannot insert \"%s\" in symbol table - invalid symbol name\n",
               token_p);
    else {
      /*{{{carry on with insertion process, checking for scoped name clashes*/
      /* look for a symbol of the same type and matching name. This can
         be ok if it was declared in a macro level less than the current one*/
      oldsym_p = look_for(token_p, any_symbol);
      if ((oldsym_p != (symtab_t*) NULL) && (oldsym_p->depth >= macro_depth))
        print("Cannot insert \"%s\" in symbol table - name clash within current scope\n", 
                token_p);
      else {
        symbol_p = (symtab_t*) malloc(sizeof(symtab_t));
        symbol_p->name_p = malloc(max_tok_len);
        strcpy(symbol_p->name_p, token_p);
        symbol_p->name_len = strlen(token_p);
        symbol_p->type = type;
        symbol_p->depth = macro_depth;
      
        /* insert new structure in table and warn if near full */
        symtab_p[symbol_cnt] = symbol_p;
        symbol_cnt++;
        if (symbol_cnt >= (max_symbol_cnt - 10))
          print("Warning: Symbol table nearly full - (%ld of %ld entries)\n",
                   symbol_cnt, max_symbol_cnt);
        }
      /*}}}*/
    }
  }
  return(symbol_p);
}
/*}}}*/
/*{{{delete_symbol*/
boolean delete_symbol(char *token_p)
/* allows deletion of a symbol from table, providing its been declared
   at the current macro_depth */
{
  symtab_t   *symbol_p;
  boolean    error = false;

  symbol_p = look_for(token_p, 0xff); /* look for any type */
  if (symbol_p == (symtab_t*)NULL)
    error = true;
  else {
    if ((symbol_p->fixed) || (symbol_p->depth != macro_depth))
      error = true;
    else {
      /* delete macro line buffers */
      if (symbol_p->type == mac_symbol) {
        macro_t  *l_p,*nl_p;
        
        l_p = symbol_p->value.mac_val;
        while (l_p != (macro_t*)NULL) {
          nl_p = l_p->line_p;
          free((char*)l_p);
          l_p = nl_p;
        }
      }
      /* mark symbol as unused, ready for purge when nest level is stripped */
      symbol_p->type = 0;
    }
  }
  return(error);
}
/*}}}*/
/*{{{assign_integer*/
boolean assign_integer(char *token_p, long value, boolean constant)
/* creates or updates an integer symbol table entry */
{
  symtab_t   *symbol_p;
  boolean    error = false;

  /* attempt to find existing symbol with matching name. If it
     doesnt exist, then create symbol and assign value. A null
     symbol name will abort the operation */
  
  if (strlen(token_p) != 0) {
    symbol_p = look_for(token_p, int_symbol);
    if ((symbol_p == (symtab_t*) NULL) && (token_p[0] != '\0')) {
      symbol_p = insert_symbol(token_p, int_symbol);
      if (symbol_p == (symtab_t*) NULL) 
        error = true;
      else {
        symbol_p->fixed = constant;
        if (symbol_p->fixed)
          symbol_p->info_p = "integer constant";
        else
          symbol_p->info_p = "integer variable";
        symbol_p->value.int_val = value;
      }
    }
    else {
      if (symbol_p->fixed)
        error = true;
      else
        symbol_p->value.int_val = value;
    }
  }
  return(error); 
}
/*}}}*/
/*{{{assign_float*/
boolean assign_float(char *token_p, double value, boolean constant)
/* creates or updates a floating point symbol table entry */
{
  symtab_t   *symbol_p;
  boolean    error = false;

  /* attempt to find existing symbol with matching name. If it
     doesnt exist, then create symbol and assign value */

  if (strlen(token_p) != 0) {
    symbol_p = look_for(token_p, flt_symbol);
    if ((symbol_p == (symtab_t*) NULL) && (token_p[0] != '\0')) {
      symbol_p = insert_symbol(token_p, flt_symbol);
      if (symbol_p == (symtab_t*) NULL) 
        error = true;
      else {
        symbol_p->fixed = constant;
        if (symbol_p->fixed)
          symbol_p->info_p = "floating point constant";
        else
          symbol_p->info_p = "floating point variable";
        symbol_p->value.flt_val = value;
      }
    }
    else {
      if (symbol_p->fixed)
        error = true;
      else
        symbol_p->value.flt_val = value;
    }
  }
  return(error); 
}
/*}}}*/
/*{{{assign_string*/
boolean assign_string(char *token_p, char *value, boolean constant)
/* creates or updates a string symbol table entry */
{
  symtab_t   *symbol_p;
  boolean    error = false;

  /* attempt to find existing symbol with matching name. If it
     doesnt exist, then create symbol and assign value */

  if (strlen(token_p) != 0) {
    symbol_p = look_for(token_p, str_symbol);
    if (symbol_p == (symtab_t*) NULL) {
      symbol_p = insert_symbol(token_p, str_symbol);
      if (symbol_p == (symtab_t*) NULL) 
        error = true;
      else {
        symbol_p->fixed = constant;
        if (symbol_p->fixed)
          symbol_p->info_p = "string constant";
        else
          symbol_p->info_p = "string variable";
        symbol_p->value.str_val = malloc(max_tok_len);
        strcpy(symbol_p->value.str_val, value);
      }
    }
    else {
      if (symbol_p->fixed)
        error = true;
      else
        strcpy(symbol_p->value.str_val, value);
    }
  }
  return(error); 
}
/*}}}*/
/*{{{register_command*/
boolean register_command(char *token_p, boolean (*action)(parse_t*, char*),
                         char *help_p)
/* establishes the existence of a command procedure */
{
  symtab_t   *symbol_p;
  boolean    error = false;

  /* attempt to find existing symbol with matching name. If it
     doesnt exist, then create symbol and assign info */

  symbol_p = look_for(token_p, (com_symbol | mac_symbol));
  if (symbol_p != (symtab_t*) NULL) {
    error = true;
    print("Name clash when registering command \"%s\"\n",token_p);
  }
  else {
    symbol_p = insert_symbol(token_p, com_symbol);
    if (symbol_p == (symtab_t*) NULL)
      error = true;
    else {
      symbol_p->fixed = true;
      symbol_p->info_p = help_p;
      symbol_p->value.com_val = action;
    }
  }
  return(error); 
}
/*}}}*/
/*{{{evaluate_integer*/
boolean evaluate_integer_expr(parse_t*, long*, short); /* forward decln */

boolean evaluate_integer(char *token_p, long *value_p, short default_base)
/* attempts a series of operations to extract an integer value from a token */
{
  parse_t  pars;
  symtab_t *symbol_p;
  boolean  error = false;

  /* eliminate null strings and then look for unary operators */
  /*print("integer evaluation of \"%s\" (in base %d)\n",token_p, 
         default_base);*/

  if (token_p[0] == '\0') error = true;
  else if (token_p[0] == '~') { 
     token_p++;
     error = evaluate_integer(token_p, value_p, default_base);
     *value_p = ~(*value_p);
  }
  else if (token_p[0] == '!') { 
     token_p++;
     error = evaluate_integer(token_p, value_p, default_base);
     *value_p = !(*value_p);
  }
  else {
    /*{{{Try to convert number or expression*/
    /*   First look for a simple number, then try a symbol reference. If this
         all fails, the value may be due to an expression, so try and evaluate it*/
    *value_p = conv_int(token_p, default_base);
    if (*value_p == bad_integer_val) {
      symbol_p = look_for(token_p, int_symbol);
      if (symbol_p == (symtab_t*) NULL) {
        init_pars(&pars, token_p);
        error = evaluate_integer_expr(&pars, value_p, default_base);
      }
      else {
        error = false;
        *value_p = symbol_p->value.int_val; 
      }
    }
    /*}}}*/
  }
  return(error);
}
/*}}}*/
/*{{{evaluate_integer_expr*/
boolean evaluate_integer_expr(parse_t *pars_p, long *value_p, 
                              short default_base)
/* generalised recursive evaluate of a possibly bracketed integer expression.
   No leading spaces are allowed in the passed string, and the parsing
   structure is assumed to be initialised but not yet used */ 
{
  boolean  error = false;
  short    brackets = 0;
  long     value1, value2;
  short    index, i;
  char     sub_expr[max_tok_len];
  char     operation;

  static   char *delim = "*/+-|&^%";   /* recognised arithmetic operators */

  /* pair up a leading bracket */
  if (pars_p->line_p[0] == '(') {
    brackets = 1;
    index = 1;
    while((pars_p->line_p[index] != '\0') && (brackets > 0)) {
      if (pars_p->line_p[index] == ')') brackets--;
      if (pars_p->line_p[index] == '(') brackets++;
      index++;
    }
    if (brackets != 0) error = true;
    else {
      /* copy substring without enclosing brackets and evaluate it.
         if this is ok, then update parsing start position */
      for(i=1; i<index; i++) sub_expr[i-1] = pars_p->line_p[i];
      sub_expr[index-2] = '\0';
      error = evaluate_integer(sub_expr, &value1, default_base);
      if (!error) pars_p->par_pos = index-1; 
    }
  }   
  if (!error) {
    /* look for a token and check for significance */
    get_tok(pars_p, delim);
    if (pars_p->tok_del == '\0') { /* no operator seen*/
      if (pars_p->par_sta > 0)       /* bracket removal code used */
        *value_p = value1;   
      else
        error = true;                /* not a valid expression! */ 
    }
    else {
      /* have found an operator. If we stripped brackets in the first
         half of this code, then the interesting part of the line is
         now after the operator. If we did not, then evaluate both
         sides of operator */
      operation = pars_p->tok_del;
      if (pars_p->par_sta == 0) 
        error = evaluate_integer(pars_p->token, &value1, default_base);
      if (!error) {
        get_tok(pars_p, "");   /* all the way to the end of this line */
        error = evaluate_integer(pars_p->token, &value2, default_base);
      }
      if (!error) {
        switch(operation) {
          case '+': *value_p = value1 + value2; break; 
          case '-': *value_p = value1 - value2; break; 
          case '*': *value_p = value1 * value2; break; 
          case '/': *value_p = value1 / value2; break; 
          case '&': *value_p = value1 & value2; break; 
          case '|': *value_p = value1 | value2; break; 
          case '^': *value_p = value1 ^ value2; break; 
          case '%': *value_p = value1 % value2; break; 
          default: error = true; break;
        }
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{evaluate_float*/
boolean evaluate_float_expr(parse_t*, double*); /* forward decln */

boolean evaluate_float(char *token_p, double *value_p)
/* attempts a series of operations to extract a floating pt value from a tok */
{
  parse_t  pars;
  symtab_t *symbol_p;
  boolean  error = false;

  /* this will only be relevant for non null strings w/o leading white space.
     First look for a simple number, then try a symbol reference. If this
     all fails, the value may be due to an expression, so try and evaluate it*/
  /* print("float evaluation of %s \n",token_p); */
  if (token_p[0] == '\0') error = true;
  else {
    error = conv_flt(token_p, value_p);
    if (error) {
      /* a symbol would be legal if it were an integer, but have to float it */ 
      symbol_p = look_for(token_p, flt_symbol | int_symbol);
      if (symbol_p == (symtab_t*) NULL) {
        init_pars(&pars, token_p);
        error = evaluate_float_expr(&pars, value_p);
      }
      else {
        error = false;
        if (symbol_p->type == flt_symbol)
          *value_p = symbol_p->value.flt_val;
        else
          *value_p = (double) symbol_p->value.int_val;
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{evaluate_float_expr*/
boolean evaluate_float_expr(parse_t *pars_p, double *value_p) 
                      
/* generalised recursive evaluate of a, possibly bracketed, float expression.
   No leading spaces are allowed in the passed string, and the parsing
   structure is assumed to be initialised but not yet used */ 
{
  boolean  error = false;
  short    brackets = 0;
  double   value1, value2;
  short    index, i;
  char     sub_expr[max_tok_len];
  char     operation;

  static   char *delim = "*/+-";   /* recognised arithmetic operators */

  /* pair up a leading bracket */
  if (pars_p->line_p[0] == '(') {
    brackets = 1;
    index = 1;
    while((pars_p->line_p[index] != '\0') && (brackets > 0)) {
      if (pars_p->line_p[index] == ')') brackets--;
      if (pars_p->line_p[index] == '(') brackets++;
      index++;
    }
    if (brackets != 0) error = true;
    else {
      /* copy substring without enclosing brackets and evaluate it.
         if this is ok, then update parsing start position */
      for(i=1; i<index; i++) sub_expr[i-1] = pars_p->line_p[i];
      sub_expr[index-2] = '\0';
      error = evaluate_float(sub_expr, &value1);
      if (!error) pars_p->par_pos = index-1; 
    }
  }   
  if (!error) {
    /* look for a token and check for significance */
    get_tok(pars_p, delim);
    if (pars_p->tok_del == '\0') { /* no operator seen*/
      if (pars_p->par_sta > 0)       /* bracket removal code used */
        *value_p = value1;   
      else
        error = true;                /* not a valid expression! */ 
    }
    else {
      /* have found an operator. If we stripped brackets in the first
         half of this code, then the interesting part of the line is
         now after the operator. If we did not, then evaluate both
         sides of operator */
      operation = pars_p->tok_del;
      if (pars_p->par_sta == 0) 
        error = evaluate_float(pars_p->token, &value1);
      if (!error) {
        get_tok(pars_p, "");   /* all the way to the end of this line */
        error = evaluate_float(pars_p->token, &value2);
      }
      if (!error) {
        switch(operation) {
          case '+': *value_p = value1 + value2; break; 
          case '-': *value_p = value1 - value2; break; 
          case '*': *value_p = value1 * value2; break; 
          case '/': *value_p = value1 / value2; break; 
          default: error = true;
        }
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{evaluate_comparison*/
boolean evaluate_comparison(char *token_p, boolean *result_p,
                            short default_base) 

/* specific evaluation of comparison between two numeric expressions.
   No leading spaces are allowed in the passed string.
   A single expression will yield a result of false if equal to zero.
   Although real expressions are allowed, comparison is done in integers.
   A default result of true is returned in the case of an error */ 
{
  boolean  error = false;
  parse_t  pars;
  long     value1, value2;
  double   real_val;
  char     op1,op2 = 0;

  static   char *delim = "><=!";   /* recognised logical operators */

  /* evaluate first arithmetic expression in line */
  *result_p = true;
  init_pars(&pars, token_p);
  get_tok(&pars, delim);
  error = evaluate_integer(pars.token, &value1, default_base);
  if (error) {
    error = evaluate_float(pars.token, &real_val);
    if (!error) value1 = (long) real_val;
  }
  if (!error) {
    
    /* deal with the case of a single expression value */
    if (pars.tok_del == '\0')
      *result_p = !(value1 == 0);
    else {

      /* look for valid single or pair of operators and move posn accordingly */
      op1 = pars.line_p[pars.par_pos];
      op2 = pars.line_p[pars.par_pos+1];
      if ((op2 != '>') && (op2 != '<') && (op2 != '=')) op2 = 0;
      else pars.par_pos++;
      
      /* get token for rest of line and extract a value in a similar way */
      get_tok(&pars, "");
      error = evaluate_integer(pars.token, &value2, default_base);
      if (error) {
        error = evaluate_float(pars.token, &real_val);
        if (!error) value2 = (long) real_val;
      }
      if (!error) {
    
        /* deal with the combination of two expression values */
        switch(op1) {
          case '=': switch(op2) {
                      case '=':  *result_p = (value1 == value2); break;
                      case '!':  *result_p = (value1 != value2); break;
                      case '<':  *result_p = (value1 <= value2); break;
                      case '>':  *result_p = (value1 >= value2); break;
                      default :  error = true; break;
                    } break;
          case '!': switch(op2) {
                      case '=':  *result_p = (value1 != value2); break;
                      default :  error = true; break;
                    } break;
          case '>': switch(op2) {
                      case '=':  *result_p = (value1 >= value2); break;
                      case  0 :  *result_p = (value1 > value2);  break;
                      default :  error = true; break;
                    } break;
          case '<': switch(op2) {
                      case '=':  *result_p = (value1 <= value2); break;
                      case  0 :  *result_p = (value1 < value2);  break;
                      default :  error = true; break;
                    } break;

        }
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{evaluate_string*/
boolean evaluate_string(char *token_p, char *string_p, short max_len)
/* attempts a series of operations to extract a string value from a token */
{
  symtab_t *symbol_p;
  boolean  error = false;
  short    index,i,len;
  short    quotes;
  parse_t  pars;
  
  /* this will only be relevant for non null strings w/o leading white space
     and enclosed in a pair of double quotes '"'. Concatenation of strings
     is dealt with using recursion*/
  
  string_p[0] = '\0';
  /* print("string evaluation of %s \n",token_p); */
  if (token_p[0] == '\0') error = true;
  else {
    /* look for concatenation function */
    init_pars(&pars, token_p);
    get_tok(&pars,"+");
    if (pars.tok_del == '+') {
      /* call routine on halves of passed token */
      error = evaluate_string(pars.token, string_p, max_len);
      if (!error) {
        get_tok(&pars, "");  /* rest of line */
        len = strlen(string_p);
        error = evaluate_string(pars.token, &string_p[len], max_len-len);
      } 
    }
    else {
      /*{{{examine given token for string contents*/
      symbol_p = look_for(token_p, str_symbol);
      if (symbol_p != (symtab_t*)NULL)
        strcpy(string_p, symbol_p->value.str_val);
      else {
        /*{{{try to extract quoted character sequence*/
        if (token_p[0] == '"') {
          quotes = 1;
          index = 1;
          while((token_p[index] != '\0') && (quotes > 0)) {
            if ((token_p[index] == '"') && (token_p[index-1] != '\\')) quotes--;
            index++;
          }
          if (quotes != 0) error = true;
          else {
            /* copy substring without enclosing quotes, preserving case */ 
            for(i=1; (i<index) && (i< max_len); i++) string_p[i-1] = token_p[i];
            string_p[i-2] = '\0';
            len = index-2;
          }
        }
        else
          error = true;
        /*}}}*/
      }
      /*}}}*/
    }
  } 
  return(error);
}
/*}}}*/
/*{{{evaluate_assign*/
boolean evaluate_assign(char *target_p, char *token_p, boolean* error_p)
/* attempts multiple evaluations of a token. If successful, assigns a 
   symbol of appropriate type to the resultant value . Two retuns are
   provided. 'error' signififies that the token was in itself
   correct, but the assignment to the target symbol failed. 'not_done'
   is used to flag an unsuccessful evaluation*/
{
  boolean not_done;
  long    int_val;
  double  real_val;
  char    string_val[max_tok_len];

  *error_p = false;
  if (strlen(target_p) != 0) {
    not_done = evaluate_integer(token_p, &int_val, number_base);
    if (!not_done) {
      *error_p = assign_integer(target_p, int_val, false);
    }
    else {
      not_done = evaluate_float(token_p, &real_val);
      if (!not_done) {
        *error_p = assign_float(target_p, real_val, false);
      }
      else {
        not_done = evaluate_string(token_p, string_val, max_tok_len);
        if (!not_done) {  
          *error_p = assign_string(target_p, string_val, false);
        }
      }
    }
  }
  return(not_done);
}
/*}}}*/
/*{{{cget_string*/
boolean cget_string(parse_t *pars_p, char *default_p, char *result_p, 
                                                     short max_len)
/* a high level parsing routine which forms part of a set. The general
   specification is the incremental reading of an input string via the
   parsing structure, detecting a parameter refernce of the right type and
   passing back a result. If the input line at this point is blank, a default
   value is substituted for a result. If the next token on the input line
   does not have the right characteristics for a parameter of this type, then
   the routine returns true to signify an error. The default parameter is
   passed back in this case as well as an error. */

{
  boolean  error = false;

  get_tok(pars_p, delim_set);
  if (pars_p->tok_len == 0)
    strncpy(result_p, default_p, max_len);
  else {
    error = evaluate_string(pars_p->token, result_p, max_len);
    if (error)
      strncpy(result_p, default_p, max_len);
  }
  return(error);
}
/*}}}*/
/*{{{cget_integer*/
boolean cget_integer(parse_t *pars_p, long def_int, long *result_p)
/* a high level parsing routine which forms part of a set. The general
   specification is the incremental reading of an input string via the
   parsing structure, detecting a parameter refernce of the right type and
   passing back a result. If the input line at this point is blank, a default
   value is substituted for a result. If the next token on the input line
   does not have the right characteristics for a parameter of this type, then
   the routine returns true to signify an error. The default parameter is
   passed back in this case as well. */

{
  boolean  error = false;

  get_tok(pars_p, delim_set);
  if (pars_p->tok_len == 0)
    *result_p = def_int;
  else {
    error = evaluate_integer(pars_p->token, result_p, number_base);
    if (error)
      *result_p = def_int;
  }
  return(error);
}
/*}}}*/
/*{{{cget_float*/
boolean cget_float(parse_t *pars_p, double def_flt, double *result_p)
/* a high level parsing routine which forms part of a set. The general
   specification is the incremental reading of an input string via the
   parsing structure, detecting a parameter refernce of the right type and
   passing back a result. If the input line at this point is blank, a default
   value is substituted for a result. If the next token on the input line
   does not have the right characteristics for a parameter of this type, then
   the routine returns true to signify an error. The default parameter is
   passed back in this case as well. */

{
  boolean  error = false;

  get_tok(pars_p, delim_set);
  if (pars_p->tok_len == 0)
    *result_p = def_flt;
  else {
    error = evaluate_float(pars_p->token, result_p);
    if (error)
      *result_p = def_flt;
  }
  return(error);
}
/*}}}*/
/*{{{cget_item*/
boolean cget_item(parse_t *pars_p, char *default_p, char *result_p, short max_len)
/* a high level parsing routine which forms part of a set. The general
   specification is the incremental reading of an input string via the
   parsing structure, detecting a parameter refernce of the right type and
   passing back a result. If the input line at this point is blank, a default
   value is substituted for a result. If the next token on the input line
   does not have the right characteristics for a parameter of this type, then
   the routine returns true to signify an error. The default parameter is
   passed back in this case as well. */

{
  boolean  error = false;

  /* simply returns next token on line or the default string */
  get_tok(pars_p, delim_set);
  if (pars_p->tok_len == 0)
    strncpy(result_p, default_p, max_len);
  else
    strncpy(result_p, pars_p->token, max_len);

  
  return(error);
}
/*}}}*/
/*{{{read_input*/
extern int io_read(char* prompt, char* buffer, long buflen);

boolean read_input(char *line_p, char *ip_prompt_p)
/* obtains input line from current active source, returns false if OK 
   SHould this routine do echo of input lines?
   should this routine issue a prompt for interactive input? */
{
  int cnt =0;
  int   temp = EOF;

  /* take input from a macro structure if it is defined, else from
     the current stream */
  if (cur_macro_p == (macro_t*) NULL) {
    if (cur_stream_p == (FILE*)NULL)
      print("asked to read from NULL file \n\n\n");
    else {
       if (cur_stream_p == stdin) {
          /* if we are at 'console level' use a user supplied input routine */
          temp = 0;
          cnt = io_read(ip_prompt_p, line_p, 255);
       }
#ifndef UNHOSTED
       else {
          /* assume reading from an input file, possibly nested */
          temp = 0; 
          line_p[cnt] = getc(cur_stream_p);
          while ((cnt < 255) && (line_p[cnt] != '\n') && 
                 ((temp = getc(cur_stream_p)) != EOF)) {
			line_p[++cnt] = temp;  
	  }
          line_p[cnt] = '\0';
       }
       /* log input line if from a file or interactive input */
       if (logfiled != NULL) 
         fprintf(logfiled, "%s\n", line_p);
#endif
    }
  }
  else {
    temp = (cur_macro_p->line_p == (macro_t*) NULL);
    strcpy(line_p, cur_macro_p->line);
    cur_macro_p = cur_macro_p->line_p;
  }
  if (cur_echo)
    print("-> %s\n", line_p);
  return(temp == EOF);
}
/*}}}*/
/*{{{define_macro*/
boolean define_macro(char *defline_p, char *name_p)

/* accepts a sequence of input lines that form the body of a macro. The
   definition line passed in contains the formal parameter part of the 
   command line that triggered the definition and may be null. Macro
   definition is ended when the first token on a line is found to be 'END' */
{
  symtab_t   *symbol_p;
  macro_t    *cur_line_p, *last_line_p;
  parse_t    pars;
  boolean    end = false;
  boolean    error = false;
  short      temp;
  short      nests = 1;

  /* attempt to find existing macro with matching name. If it
     doesnt exist, then create symbol and set up parameter string */

  symbol_p = look_for(name_p, any_symbol);
  if ((symbol_p == (symtab_t*) NULL) && (name_p[0] != '\0')) {
    symbol_p = insert_symbol(name_p, mac_symbol);
    if (symbol_p == (symtab_t*) NULL) { 
      print("Cannot create macro \"%s\" \n", name_p);
      error = true;
    }
    else {
      symbol_p->fixed = false;
      symbol_p->info_p = "command macro";
      symbol_p->value.mac_val = (macro_t*) malloc(sizeof(macro_t));
      strcpy(symbol_p->value.mac_val->line, defline_p);
      last_line_p = symbol_p->value.mac_val;
      while (!end) {
        cur_line_p = (macro_t*) malloc(sizeof(macro_t));
        last_line_p->line_p = cur_line_p;
        last_line_p = cur_line_p;
        read_input(cur_line_p->line, "> ");
        init_pars(&pars, cur_line_p->line);
        get_tok(&pars, delim_set);
        if ((is_matched(pars.token, "END", 2)) ||
            (is_matched(pars.token, "EXIT", 2))) {
          nests--;
          if (nests == 0) end = true;
        }
        else if (is_control(pars.token, &temp)) nests++;
      }
      last_line_p->line_p = (macro_t*)NULL;
    }
  }
  else {
    print("Cannot define macro \"%s\" - name clash\n", name_p);
    error = true;
  }
  return(error);
} 
/*}}}*/
/*{{{execute_command_line*/
boolean execute_command_line(char *line_p, char *target_p, boolean *result_p)
/* this routine assumes a string is passed which contains only a
   potential command name followed by a set of parameters. If the first
   token on the line is not a valid command, the routine returns true. If
   it is, then the defined action routine for that command is invoked and the
   result of the execution passed back to the caller */
{
   symtab_t   *command_p;
   parse_t    pars;
   boolean    error;

   error = false;
   init_pars(&pars, line_p);
   get_tok(&pars, delim_set);
   command_p = look_for(pars.token, com_symbol);
   if (command_p != (symtab_t*) NULL) 
     *result_p = command_p->value.com_val(&pars, target_p);
   else
     error = true;
   return(error);
}
/*}}}*/
/*{{{execute_macro*/
boolean command_loop(macro_t* , FILE*, char*, boolean); /* forward definition */

boolean execute_macro(char *line_p, char *target_p, boolean *result_p)
/* this routine assumes a string is passed which contains only a
   potential macro name followed by a set of actual parameters. 
   If the first token on the line is not a current macro, the routine returns 
   true. If it is, then the body of the macro is executed and the
   result of the execution passed back to the caller. If the macro attempts
   to pass back a value, the expression is evaluated into the target symbol 
   name */
{
   symtab_t   *macro_p;
   parse_t    formal_pars, actual_pars;
   boolean    error = false;
   boolean    not_done = false;
   char       rtn_expression[max_tok_len];
   long       int_val;
   double     real_val;
   char       string_val[max_tok_len];
   short      type;

   init_pars(&actual_pars, line_p);
   get_tok(&actual_pars, delim_set);
   macro_p = look_for(actual_pars.token, mac_symbol);
   if (macro_p != (symtab_t*) NULL) {
     /* define the formal parameters with values of actuals. look over
     the length of the invocation line and pick up the actual parameters
     determine their type and create an equivalent symbolic variable
     under the name of the formal parameter at the same position. formal
     parameters with no actual values are given default types and values. */

     macro_depth++;
     init_pars(&formal_pars, macro_p->value.mac_val->line);
     get_tok(&formal_pars, delim_set);
     while(( formal_pars.tok_len > 0) && !error) {
       get_tok(&actual_pars, delim_set);
       if (actual_pars.tok_len > 0) {
         not_done = evaluate_assign(formal_pars.token, actual_pars.token, &error);
       }
       else not_done = evaluate_assign(formal_pars.token, "0", &error);
       get_tok(&formal_pars, delim_set);
     }
     if (error || not_done) 
       tag_current_line(&actual_pars, 
             "Failed to assign actual value to formal parameter");
     else {
       /* if all this parameterisation worked out OK, then actually execute the
       macro code and return the result. If an an expression was found at the
       end of the macro we must evaluate it into a temporary variable in the
       scope of the macro and then reassign it to the target variable in the
       enclosing scope */

       *result_p = command_loop(macro_p->value.mac_val->line_p,
                                           (FILE*) NULL, rtn_expression, false);
       /*{{{sort out type of return expression and evaluate it*/
       if (strlen(rtn_expression) != 0) { 
         *result_p = evaluate_integer(rtn_expression, &int_val, number_base);
         if (!*result_p)
           type = int_symbol;
         else {
           *result_p = evaluate_float(rtn_expression, &real_val);
           if (!*result_p) 
             type = flt_symbol;
           else { 
             *result_p = evaluate_string(rtn_expression, string_val, max_tok_len);
             if (!*result_p)
               type = str_symbol;
             else
               type = 0;
           }
         }
       }
       /*}}}*/
     }
     macro_depth--;
     purge_symbols(macro_depth);
     /* now we are back within original scope, try to evaluate and assign
        return expression, using temporary variable to give value*/
     if ((strlen(rtn_expression) != 0) && (type != 0)) {
       switch(type) {
         case int_symbol: *result_p = assign_integer(target_p, int_val, false); break;
         case flt_symbol: *result_p = assign_float(target_p, real_val, false); break;
         case str_symbol: *result_p = assign_string(target_p, string_val, false); break;
       }
     }
   }
   else
     error = true;
   return(error);
}

/*}}}*/
/*{{{do_show*/
boolean do_show(parse_t *pars_p, char *result_sym_p)
/* performs the actions of displaying constants and symbols. If a
   line of parameters is present, each is interpreted and displayed. Otherwise
   the whole current list of symbols is displayed. */

/* This current version amounts to a debug tool and needs some polishing */
{
  boolean   error = false;
  symtab_t  *symbol_p;
  short     cnt = 0;
  macro_t   *line_p;
  

  if (get_tok(pars_p, delim_set) == 0) {  /* no items on a line */
    print("Currently defined symbols:\n\n");
    /*{{{print symbol table info*/
    while(cnt < symbol_cnt) {
      symbol_p = symtab_p[cnt++];
      switch(symbol_p->type) {
        case int_symbol:print("%2d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("%ld - %s\n", symbol_p->value.int_val,
                                              symbol_p->info_p); 
                        break;
        case flt_symbol:print("%2d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("%G - %s\n", symbol_p->value.flt_val,
                                              symbol_p->info_p); 
                        break;
        case str_symbol:print("%2d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("\"%s\" - %s\n", symbol_p->value.str_val,
                                              symbol_p->info_p); 
                        break;
        case mac_symbol:print("%2d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("(%s) - %s\n", symbol_p->value.mac_val->line,
                                              symbol_p->info_p); 
                        break;
      }
    }
    /*}}}*/
  }
  else {
    symbol_p = look_for(pars_p->token, any_symbol);
    if (symbol_p == NULL)
      tag_current_line(pars_p, "unrecognised symbol name");
    else {
      switch(symbol_p->type) {
        case int_symbol:print("%d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("%ld (#%x) - %s\n", symbol_p->value.int_val,
                                              symbol_p->value.int_val,
                                              symbol_p->info_p); 
                        break;
        case flt_symbol:print("%d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("%G - %s\n", symbol_p->value.flt_val,
                                              symbol_p->info_p); 
                        break;
        case str_symbol:print("%d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("\"%s\" - %s\n", symbol_p->value.str_val,
                                              symbol_p->info_p); 
                        break;
        case com_symbol:print("%d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("is a command\n"); 
                        break;
        case mac_symbol:print("%d: %s :",symbol_p->depth, symbol_p->name_p);
                        print("(%s) - %s\n", symbol_p->value.mac_val->line,
                                              symbol_p->info_p); 
                        /*{{{print lines of macro*/
                        /* print macro contents */
                        line_p = symbol_p->value.mac_val->line_p;
                        while(line_p != NULL) {
                           print(">   %s\n", line_p->line);
                           line_p = line_p->line_p;
                        }
                        /*}}}*/
                        break;
        default:        print("%d: %s has been deleted\n",symbol_p->depth, 
                                            symbol_p->name_p);
                        break;
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{do_print*/
boolean do_print(parse_t *pars_p, char *result_sym_p)
/* perform a fomatting action on a list of variables or values by
   invoking the sprintf routine */
{
  boolean   error = false;
  short     cnt;
  long      int_val;
  double    real_val;
  char      str_val[max_tok_len];
  char      str_buf[255];
  
  str_buf[0] = '\0';
  cnt = 0;
  while ((pars_p->tok_del != '\0') && (strlen(str_buf) < 200)) {
    get_tok(pars_p, delim_set);
    error = evaluate_integer(pars_p->token, &int_val, number_base);
    if (!error) {
      switch (number_base) {
        case  2: {
                   int i;
                   str_val[0]='$';
                   str_val[33] = '\0';
                   for (i=31; i>=0; i--)
                   if ( ((int_val>>i) & 0x1) > 0)
                     str_val[32-i] = '1';
                   else
                     str_val[32-i] = '0';
                 } break;
        case  8: sprintf(str_val, "O%lo", int_val); break;
        case 10: sprintf(str_val, "%ld", int_val); break;
        case 16: sprintf(str_val, "#%lx", int_val); break;
      }
      strcat(str_buf, str_val);
    }
    else {
      error = evaluate_float(pars_p->token, &real_val);
      if (!error) {
        sprintf(str_val, "%G", real_val);
        strcat(str_buf, str_val);
      }
      else {
        error = evaluate_string(pars_p->token, str_val, max_tok_len);
        if (!error) 
          strcat(str_buf, str_val);
        else
          strcat(str_buf, "******");
      }
    }
    cnt++;
  }
  print("%s\n", str_buf);

  return(error || assign_string(result_sym_p, str_buf, false));
}
/*}}}*/
/*{{{do_help*/
boolean do_help(parse_t *pars_p, char *result_sym_p)
/* Allows display of help information attached to each command or macro
   currently defined. If a parameter is present, the display is limited to
   help for that command. Otherwise a list is printed. */
{
  boolean   error = false;
  symtab_t  *symbol_p;
  short     cnt = 0;

  if (get_tok(pars_p, delim_set) == 0) {  /* no items on a line */
    /*{{{display list of commands and macros plus general help*/
    print("The following commands and macros are defined:\n\n");
    while(cnt < symbol_cnt) {
      symbol_p = symtab_p[cnt++];
      switch(symbol_p->type) {
        case com_symbol:
          print("%s - %s\n", symbol_p->name_p, symbol_p->info_p); 
          break;
        case mac_symbol:
          print("%s ", symbol_p->name_p);  
          print("(%s) - %s\n", symbol_p->value.mac_val->line, symbol_p->info_p);
          break;
        default: break;
      }
    }
    print("\n");
    print("Command and macros can be used as statements or part of expressions,\n");
    print("if appropriate, where the assignment is made to a symbol whose value\n");
    print("is either defined or modified by this assignment. Generalised assignments\n");
    print("can be made to symbols of integer, floating point or string types,\n");
    print("including arbitrary expression evaluation.\n\n");
    print("Commands and assignments can be combined into parameterised macros\n");
    print("that use FOR, IF, ELSE and WHILE constructs. Using the DEFINE\n");
    print("command named macros can be set up for later invocation. All blocks\n");
    print("within these constructs is terminated using the END statement.\n\n");
    
    /*}}}*/
  }
  else {
    /*{{{display individual information*/
    symbol_p = look_for(pars_p->token, (mac_symbol | com_symbol));
    if (symbol_p != (symtab_t*)NULL) {
      macro_t *l_p;
      switch(symbol_p->type) {
        case com_symbol:
          print("%s - %s\n", symbol_p->name_p, symbol_p->info_p); 
          break;
        case mac_symbol:
          print("%s ", symbol_p->name_p);  
          print("(%s) - %s\n", symbol_p->value.mac_val->line, symbol_p->info_p);
          l_p = symbol_p->value.mac_val->line_p;
          while( l_p != (macro_t*)NULL) {
            print("  %s\n", l_p->line);
            l_p = l_p->line_p;
          }
          break;
        default: break;
      }
    }
    else {
      tag_current_line(pars_p, "unrecognised command or macro");
      error = true;
    }
    /*}}}*/
  }
  return(error || assign_string(result_sym_p, "", false));
}
/*}}}*/
/*{{{do_delete*/
boolean do_delete(parse_t *pars_p, char *result_sym_p)
/* performs the action of deleting symbols and macros. If
   line of parameters is present, each is interpreted
   and, if possible, deleted. */
{
  boolean   error = false;
  symtab_t  *symbol_p;
  

  if (get_tok(pars_p, delim_set) == 0) {  /* no items on a line */
    tag_current_line(pars_p, "Expected symbol or macro name");
    error = true;
  }
  else while ((pars_p->tok_len > 0) && !error) {
    symbol_p = look_for(pars_p->token, 0xff);
    if (symbol_p == (symtab_t*)NULL) {
      tag_current_line(pars_p, "Unrecognised symbol");
      error = true;
    }
    else if (symbol_p->fixed || (symbol_p->type == com_symbol)) {
      tag_current_line(pars_p, "Cannot delete fixed symbol or command");
      error = true;
    }
    else {
      error = delete_symbol(pars_p->token);
      if (error)
        tag_current_line(pars_p, "Cannot delete symbol out of current scope");
      get_tok(pars_p, delim_set);
    }
  }
  return(error || assign_string(result_sym_p, "", false));
}
/*}}}*/
/*{{{do_verify*/
boolean do_verify(parse_t *pars_p, char *result_sym_p)
/* toggles the echo of input lines */
{
  boolean   error;
  long      echo;
  
  error = cget_integer(pars_p, cur_echo, &echo);
  if (error) 
    tag_current_line(pars_p, "expected echo flag");
  else {
    cur_echo = (boolean) echo;
    if (cur_echo)
      print("Input echo is enabled\n");
    else
      print("Input echo is disabled\n");
  }
  return(error || assign_integer(result_sym_p, echo, false));
}
/*}}}*/

/*{{{do_base*/
boolean do_base(parse_t *pars_p, char *result_sym_p)
/* performs the action of defining the default number base */
{
  boolean   error;
  long      base;
  
  error = cget_integer(pars_p, number_base, &base);
  if (error ||
      ((base != 16) && (base != 10) && (base != 8) && (base != 2))) {
    tag_current_line(pars_p, "Illegal number base");
    error = true;
  }
  else {
    number_base = (short) base;
    print("Number base = %d\n", number_base);
  }
  return(error || assign_integer(result_sym_p, number_base, false));
}
/*}}}*/
#ifndef UNHOSTED
/*{{{do_log*/
boolean do_log(parse_t *pars_p, char *result_sym_p)
/* controls the logging of input to a journal file */
{
  boolean   error = false;
  char      filename[FILENAME_MAX];
  

  if (cget_string(pars_p, "logfile.com", filename, FILENAME_MAX) ||
      (strlen(filename) == 0)) {
    tag_current_line(pars_p, "expected log file name");
    error = true;
  }
  else {
    /* close existing file, if any */
    if (logfiled != NULL)
       fclose(logfiled);
    /* attempt to open new file for writing */
    logfiled = fopen(filename, "w");
    if (logfiled != NULL) {
      print("Logging input to file \"%s\"\n",filename);
    }
    else
      print("Cannot open file \"%s\" for output\n", filename);
  }
  return(error || assign_string(result_sym_p, filename, false));
}
/*}}}*/
/*{{{do_source*/
boolean do_source(parse_t *pars_p, char *result_sym_p)
/* performs the execution of a command input file */
{
  boolean   error = false;
  FILE      *inputfd;
  char      filename[FILENAME_MAX];
  long      flag;
  

  if (cget_string(pars_p, "default.com", filename, FILENAME_MAX) ||
      (strlen(filename) == 0)) {
    tag_current_line(pars_p, "Illegal input file");
    error = true;
  }
  else {
    inputfd = fopen(filename, "r");
    if (inputfd != (FILE*)NULL) {
      print("input from file \"%s\"\n",filename);
      /* look for echo option */
      cget_integer(pars_p, 0, &flag);
      command_loop((macro_t*)NULL, inputfd, (char*)NULL,(boolean) flag);
      fclose(inputfd);
    }
    else
      print("Cannot open file \"%s\" for input\n", filename);
  }
  return(error || assign_string(result_sym_p, filename, false));
}
/*}}}*/
#endif
/*{{{do_define*/
boolean do_define(parse_t *pars_p)
/* allows definition of command macros as a control command in itself */
{
  boolean   error = false;
  char      name[max_tok_len];
  

  if (get_tok(pars_p, delim_set) == 0) {  /* attempt to find macro name */
    tag_current_line(pars_p, "macro name expected");
    error = true;
  }
  else {
    strcpy(name, pars_p->token);
    get_tok(pars_p, ""); /* get the rest of the line */
    error = define_macro(pars_p->token, name);
  }
  return(error);
}
/*}}}*/
/*{{{do_for*/
boolean do_for(parse_t *pars_p)
/* allows definition of flexible command repetition loop */
{
  boolean   error = false;
  long      first = 0;
  long      second = 0;
  long      step;
  long      i;
  char      variable[max_tok_len];
  char      name[max_tok_len];
  

  if (get_tok(pars_p, delim_set) == 0) {  /* attempt to find parameters */
    tag_current_line(pars_p, "for loop variable name expected");
    error = true;
  }
  else {
    strcpy(variable, pars_p->token);
    error = assign_integer(variable, 0, false);
    if (error)
      tag_current_line(pars_p, "expected integer loop variable name");
    else {
      error = cget_integer(pars_p, 1, &first);
      if (error)
        tag_current_line(pars_p, "invalid loop start value");
      else {
        error = cget_integer(pars_p, 1, &second);
        if (error)
          tag_current_line(pars_p, "invalid loop end value");
        else {
          /*{{{get consistent step value and execute*/
          error = (cget_integer(pars_p, 1, &step) ||
                   ((first > second) && (step > 0)) ||
                   ((first < second) && (step < 0)) ||
                   (step == 0));
          if (error) 
            tag_current_line(pars_p, "invalid or inconsistent loop step value");
          else {
            sprintf(name, "FOR%d", macro_depth);
            error = define_macro("", name);
            if (!error) {
              if (first <= second) {
                for(i=first; (i<=second) && !error; i=i+step) {
                  assign_integer(variable, i, false);
                  execute_macro(name,"",&error);
                }
              }
              else {
                for(i=first; (i>=second) && !error; i=i+step) {
                  assign_integer(variable, i, false);
                  execute_macro(name,"",&error);
                }
              }
            }
            delete_symbol(name);
          }
          /*}}}*/
        }
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{do_while*/
boolean do_while(parse_t *pars_p)
/* allows definition of conditional command repetition loop */
{
  boolean   error = false;
  boolean   result = false;
  char      name[max_tok_len];
  char      condition[max_tok_len];
  

  if (get_tok(pars_p, delim_set) == 0) {  /* attempt to find parameters */
    tag_current_line(pars_p, "comparison expression expected");
    error = true;
  }
  else {
    error = evaluate_comparison(pars_p->token, &result, number_base);
    if (error) tag_current_line(pars_p, "illegal comparison");
    else {
      sprintf(name, "WHILE%d", macro_depth);
      strcpy(condition, pars_p->token);
      error = define_macro("", name);
      if (!error) {
        while (result && !error) {
          execute_macro(name,"",&error);
          evaluate_comparison(condition, &result, number_base);
        }
        delete_symbol(name);
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{do_if*/
boolean do_if(parse_t *pars_p, boolean *if_taken_p)
/* allows definition of conditional command execution, passing the result
   of the condition evaluation back to a caller for use by the 'else' clause */
{
  boolean   error = false;
  char      name[max_tok_len];
  char      condition[max_tok_len];
  

  if (get_tok(pars_p, delim_set) == 0) {  /* attempt to find parameters */
    tag_current_line(pars_p, "comparison expression expected");
    error = true;
  }
  else {
    error = evaluate_comparison(pars_p->token, if_taken_p, number_base);
    if (error) tag_current_line(pars_p, "illegal comparison");
    else {
      sprintf(name, "IF%d", macro_depth);
      strcpy(condition, pars_p->token);
      error = define_macro("", name);
      if (!error) {
        if (*if_taken_p)
          execute_macro(name,"",&error);
        delete_symbol(name);
      }
    }
  }
  return(error);
}
/*}}}*/
/*{{{do_else*/
boolean do_else(boolean if_taken)
/* allows definition of conditional command execution after
   evaluation of the 'if condition */
{
  boolean   error = false;
  char      name[max_tok_len];
  
  sprintf(name, "ELSE%d", macro_depth);
  error = define_macro("", name);
  if (!error) {
    if (!if_taken)
      execute_macro(name,"",&error);
    delete_symbol(name);
  }
  return(error);
}
/*}}}*/
/*{{{command_loop*/
boolean command_loop(macro_t *macro_p, FILE *file_p, char *rtn_exp_p,
                                                    boolean echo)
/* fundamental command execution loop. Commands come either from
a supplied string or from an input stream (file or TTY). The loop
identifies the type of input statement and determines the action to take
as a result of this. A range of simple execution control primitives have
been supplied. Block structure is implemented using temporary macro-like
structures. */
{
  boolean  error = false;
  boolean  exit = false;
  boolean  not_done = true;
  char     input_line[255];
  boolean  sav_echo;
  macro_t  *sav_macro_p;
  FILE     *sav_stream_p; 
  parse_t  pars;
  short    control_const;
  boolean  if_flag; /* set if seen in the last line examined */
  boolean  if_taken;
  
  static char std_delim[] = "= ,\\";        /* normal command line delimiters */

  /* save the existing input source and substitute a new one */
  sav_echo = cur_echo;
  sav_macro_p = cur_macro_p;
  sav_stream_p = cur_stream_p;

  cur_echo = echo;
  cur_macro_p = macro_p;
  cur_stream_p = file_p;

  if_flag = false;

  while( !exit && (!error || (macro_depth <= 1))) {
    /* read an input line and start parsing process. This nesting level
       will terminate if the end of the stream is reached or an error
       is detected. The outermost loop, generally interactive, will not
       terminate on error. */
    exit = read_input(input_line, prompt_p); 
    init_pars(&pars, input_line);
    get_tok(&pars, std_delim);
    
    if (exit || is_matched(pars.token, "EXIT", 2)
             || is_matched(pars.token, "END", 2)) exit = true;
    else if (pars.tok_len > 0) {
      /* proceed with examination of an input line. Broadly speaking
         these come in three flavours - assignment, statement and control. The
         key distiguishing features are a leading '=' seperator or
         the first token being a member of a small set of reserved
         control words. In each executable
         case, an attempt is made to  execute the line in various
         forms before declaring an error. */
      if (pars.tok_del == '=') {
        /*{{{attempt to make sense of assigment operation */
        /* try and make an assignment work out */
        char   target[max_tok_len];
        
        if_flag = false;
        strcpy(target, pars.token);
        get_tok(&pars, "");        /* tokenise rest of line */
        not_done = evaluate_assign(target, pars.token, &error);
        if (not_done) 
        /* attempt to execute a command or macro assignment*/
          not_done = execute_command_line(pars.token, target, &error);
        if (not_done)
          not_done = execute_macro(pars.token, target, &error);
        if (not_done)
          tag_current_line(&pars, "Unrecognised assignment statement");
        /* Report an error in a separate fashion */
        if (error) print("Assignment of \"%s\" failed\n", target); 
        /*}}}*/
      }
      else if (is_control(pars.token, &control_const)) {
        /*{{{deal with a control construct*/
        switch(control_const) {
          case define_const:  
            error = do_define(&pars);
            if_flag = false; 
            break;
          case if_const:
            error = do_if(&pars, &if_taken);
            if (!error) if_flag = true;
            break;
          case else_const:
            if (!if_flag) {
              error = true;
              tag_current_line(&pars, "ELSE not allowed without IF");
            }
            else {
              error = do_else(if_taken);
              if_flag = false;
            }
            break;
          case while_const:
            error = do_while(&pars);
            if_flag = false;
            break;
          case for_const:
            error = do_for(&pars);
            if_flag = false;
            break;
          default:       
            tag_current_line(&pars, "Unable to understand control construct"); 
            error = true;
            break;
        }
        /*}}}*/
      } 
      else {
        /*{{{attempt to execute a command or macro statement*/
        if_flag = false;
        not_done = execute_command_line(input_line, "", &error);
        if (not_done)
          not_done = execute_macro(input_line, "", &error);
        if (not_done) {
          tag_current_line(&pars, "Unrecognised command statement");
          error = true;
        }
        /*}}}*/
      }
    }
  }
  
  /* if this is a normal termination of the loop via an END statement,
     look for an expression to use as a result. As usual, the type depends
     on the context of the expression, and this is evaluated in the
     scope of the enclosing macro invocation */
  if (!error) {
    if ((get_tok(&pars, delim_set) != 0) && (rtn_exp_p != (char*)NULL)) {
      strcpy(rtn_exp_p, pars.token);
    }
  }
  /* restore original input information */
  cur_echo = sav_echo;
  cur_macro_p = sav_macro_p;
  cur_stream_p = sav_stream_p;

  return(error);
}
/*}}}*/
/*{{{cli_main*/
void cli_main(boolean (*setup_r)(), long max_symbols, short default_base,
              char *ip_prompt_p, char *file_p)
/* top level of the cli code, invokes intialisation and top level shell */
{
#ifndef UNHOSTED
  FILE  *inputfd;
#endif

  macro_depth = 0;
  number_base = default_base; 
  prompt_p = ip_prompt_p;
  init_sym_table(100+max_symbols);
  
  /* set up of all internal commands and constant values */
  register_command("HELP", do_help, "Displays help for named commands and macros");
  register_command("SHOW", do_show, "Displays symbol values and macro contents");
  register_command("DELETE", do_delete, "Removes named symbols and macros");
  register_command("IOBASE", do_base, "<int> Sets default integer I/O radix");
  register_command("VERIFY", do_verify, "<flag> Sets default input echo");
  register_command("PRINT", do_print, "Formats and prints variables");

#ifndef UNHOSTED
  register_command("SOURCE", do_source, "<filestring><echo> Executes commands in file");
  register_command("LOG", do_log, "<filestring> Logs command input to file");
#endif
  assign_integer("HEXADECIMAL", 16, true);
  assign_integer("DECIMAL", 10, true);
  assign_integer("OCTAL", 8, true);
  assign_integer("BINARY", 2, true);

  assign_integer("TRUE", 1, true);
  assign_integer("FALSE", 0, false);

  assign_float("PI", 3.14156, true);

  /* call user defined setup for commands and symbols */
  (*setup_r)();

  macro_depth++;    /* increase nesting depth for initial input */
  logfiled = NULL;  /* initialise log file */

#ifndef UNHOSTED

  /* if an initial input file is specified, invoke command 
     processing on contents of that file */ 
  if ((file_p != NULL) && (strlen(file_p) != 0)) {
    inputfd = fopen(file_p, "r");
    if (inputfd != (FILE*)NULL) {
      print("input from file \"%s\"\n",file_p);
      command_loop((macro_t*)NULL, inputfd, (char*)NULL, false);
    }
    else
      print("Cannot open file \"%s\" for input\n", file_p);
  }
#endif

  /* invoke an interactive command loop */
  
  command_loop((macro_t*)NULL, stdin, (char*)NULL, false);
  
  /* purge symbol table of contents */
  macro_depth--;
  purge_symbols(macro_depth);

#ifndef UNHOSTED  
  /* if we have a log file open then close it properly */
  if (logfiled != NULL)
    fclose(logfiled);

#endif
} 
/*}}}*/
