/*{{{ COMMENT Standard Header*/
/******************************************************************
COPYRIGHT (C) INMOS LIMITED - 1990, 1991

    Source file name : basic.c
              Author : S.R.Christian, D. Johnston
   Date first issued : 12-Apr-91

Low level routines to support hardware of IMS B016


Date        Modification                                    Initials
----        ------------                                    --------
02-Oct-91   Adapted from original                           SRC
******************************************************************/

/*{{{Info*/
/*
   Although all of these subroutines may prove useful in
   different applications, the only ones which are strictly
   required outside this module are the following:

   void shadow_write_map_ram(address,data)

   writes the data to the map ram location corresponding to
   the address, as also to a shodow array so that the map
   ram can be readback.

   int shadow_read_map_ram(address)

   returns the map ram contents corresponding to the address.
   Actually reads the shadow array.

   clear_map_ram() does the obvious.

   enable_decode_ram(address) and
   disable_decode_ram(address)

   respectively enable and disable VMEbus slave decoding for
   particular address regions, by writing the decode RAM.

   clear_decode_ram() clears all decoding.

   basic_board_init()

   sets up the board is a basic condiguration of map-ram and
   clears VIC refresh. This routine also sets a flag which
   can be checked at any time by calling

   int check_board_init();   which returns zero if the board
   has not yet been initialised. Various routines use this
   flag and the routine and refuse to work if the basic_board_init()
   has not been called yet.
*/
/*}}}*/

/*}}}*/

#include "basic.h"

/*{{{Static variables*/
static short map_ram_shadow[4096];
static void *map_ram_routine;
static char map_ram_routine_status = 0;
static char board_init_status = 0;

static void *decode_ram_routine;
static char decode_ram_routine_status = 0;
/*}}}*/

void *copy_map_ram_routine(void)
/*{{{*/
{
 int i;
 int memstart;
 int retval;
 int *p;
 int hexcode[9]=
     {
     0x67D0BE60,0x20202020,0xE0402720,0x2D2F2741,
     0x2420202C,0x2067E048,0x20202020,0xE0304027,
     0x20F6B270
     };

  /*  We're using a b016 so memstart=0x80000070 always */

  memstart=0x80000070;
  p=(int *)(memstart+4);
  for(i=0;i<9;i++)           /* copy the assembly code to */
    {                       /* the internal ram.         */
      *p=hexcode[i];
      p++;
    }
  map_ram_routine_status  = 1;
  retval=memstart+4;
  return ((void *)(memstart+4));     /* returns the address */
                                       /* of the code.        */
}
/*}}}*/

void basic_map_ram(void *set_map, void *map_addr,int val)
/*{{{*/
{
        __asm
        {
                ld val;
                ld map_addr;
                ld set_map;
                gcall;
        }
}
/*}}}*/

void init_map_ram_routine(void)
/*{{{*/
{
        map_ram_routine = copy_map_ram_routine();
}
/*}}}*/

void write_map_ram(void *map_addr,int val)
/*{{{*/
{
        if (map_ram_routine_status == 0)
                init_map_ram_routine();
        basic_map_ram(map_ram_routine,map_addr,val);
}
/*}}}*/

void shadow_write_map_ram(void* address, int data)
/*{{{*/
{
        map_ram_shadow[((unsigned int)address)>>20] = data;
        write_map_ram(address,data);
}
/*}}}*/

int shadow_read_map_ram(void* address)
/*{{{*/
{
        return (map_ram_shadow[((unsigned int)address)>>20]);
}
/*}}}*/

void clear_map_ram(void)
/*{{{*/
{
        int count;
        for (count = 0; count < 4096; count++) {
                shadow_write_map_ram((void *)(((count*(1048576))+4096)),0);
        }
}
/*}}}*/

void *copy_decode_ram_routine(void)
/*{{{*/
{int i;
 int memstart;
 int *p;
 int hexcode[9]=
     {
     0x67D0BE60,0x20202020,0xE0402720,0x2D2F2741,
     0x2420202C,0x2067E04C,0x20202020,0xE0304027,
     0x20F6B270
     };

  /*  We're using a b016 so memstart=0x80000070 always */

  memstart=0x80000070;
  p=(int *)(memstart+40);   /* leave space for mapram code */
  for(i=0;i<9;i++)           /* copy the assembly code to */
    {                       /* the internal ram.         */
      *p=hexcode[i];
      p++;
    }
  decode_ram_routine_status  = 1;
  return ((void *)(memstart+40));      /* returns the address */
                                       /* of the code.        */
}
/*}}}*/

void basic_decode_ram(void *routine,void *address,int val)
/*{{{*/
{
        __asm
        {
                ld val;
                ld address;
                ld routine;
                gcall;
        }
}
/*}}}*/

void init_decode_ram_routine(void)
/*{{{*/
{
        decode_ram_routine = copy_decode_ram_routine();
}
/*}}}*/

void write_decode_ram(void *address,int val)
/*{{{*/
{
int addr2;
    if (decode_ram_routine_status == 0)
        init_decode_ram_routine();
    if (((unsigned int)address & 0xfffff000) == 0x80000000)
        {
          addr2=(int)address;
          addr2+=(int)0x8000;
          address=(void *)addr2;
        }
    basic_decode_ram(decode_ram_routine,address,val);
}
/*}}}*/

void enable_decode_ram(void *address)
/*{{{*/
{
          write_decode_ram(address,2);
}
/*}}}*/

void disable_decode_ram(void *address)
/*{{{*/
{
          write_decode_ram(address,0);
}
/*}}}*/

void clear_decode_ram(void)
/*{{{*/
{
        int count;
        for (count = 0; count < 65536; count++) {
                write_decode_ram((void *)(count*65536),0);
        }
}
/*}}}*/

int check_board_init(void)
/*{{{*/
{
        return((int) board_init_status);
}
/*}}}*/

void clear_vic(void)
/*{{{*/
{
*((uint *)VIC_INT_CONT_ENB)=0XF0;
*((uint *)VIC_VME_INT_REG1)=0XF0;
*((uint *)VIC_VME_INT_REG2)=0XF0;
*((uint *)VIC_VME_INT_REG3)=0XF0;
*((uint *)VIC_VME_INT_REG4)=0XF0;
*((uint *)VIC_VME_INT_REG5)=0XF0;
*((uint *)VIC_VME_INT_REG6)=0XF0;
*((uint *)VIC_VME_INT_REG7)=0XF0;
*((uint *)VIC_LOC_INT_REG1)=0XF0;
*((uint *)VIC_LOC_INT_REG2)=0XF0;
*((uint *)VIC_LOC_INT_REG3)=0XF0;
*((uint *)VIC_LOC_INT_REG4)=0XF0;
*((uint *)VIC_LOC_INT_REG5)=0XF0;
*((uint *)VIC_LOC_INT_REG6)=0XF0;
*((uint *)VIC_LOC_INT_REG7)=0xF0;
}
/*}}}*/

void basic_board_init(void)
/*{{{*/
{
  clear_decode_ram();
  clear_map_ram();
  shadow_write_map_ram((void *)VIC_RESET_REG,(int)LADD11);
  clear_vic();

  *(unsigned int *)VIC_REQUEST_REG = 0x00; /* turn off VIC memory refresh */
  *(unsigned int *)SPLAT_EV = 0x100; /* turn off VIC memory refresh */

  board_init_status = 1;
}
/*}}}*/

void turnon_byte_mux(void)
/*{{{*/
{
  *((unsigned int *)ENABLE_BYTE_MUX) = 1;
}
/*}}}*/

void turnoff_byte_mux(void)
/*{{{*/
{
  *((unsigned int *)ENABLE_BYTE_MUX) = 0;
}
/*}}}*/

void program_byte_mux(int which_page,int read_swap,int write_swap)
/*{{{*/
{
int swap_function[4]={0,0,0,0};
int port;
int i;
int word;
int nybble;
  for(i=0;i<8;i++)
    {
      port=(((read_swap & 0xf)+8)&0xf); /* Gives port number for swap table */
      read_swap=read_swap >> 4;
      word=i % 4;        /* gives word into which port will go */
      nybble=(i/4)+4;    /* gives nybble of word which port goes into */
      swap_function[word]=swap_function[word] | (port << (nybble * 4));
    }
  for(i=0;i<8;i++)
    {
      port=(write_swap & 0xf); /* Gives port number for swap table */
      write_swap=write_swap >> 4;
      word=i % 4;        /* gives word into which port will go */
      nybble=(i/4)+6;    /* gives nybble of word which port goes into */
      swap_function[word]=swap_function[word] | (port << (nybble * 4));
    }

  /*{{{*/
  switch (which_page)
    {
      case 0:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE0_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE0_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE0_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE0_3) = swap_function[3];
               break;
             }
             /*}}}*/
      case 1:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE1_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE1_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE1_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE1_3) = swap_function[3];
               break;
             }
             /*}}}*/
      case 2:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE2_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE2_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE2_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE2_3) = swap_function[3];
               break;
             }
             /*}}}*/
      case 3:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE3_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE3_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE3_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE3_3) = swap_function[3];
               break;
             }
             /*}}}*/
      case 4:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE4_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE4_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE4_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE4_3) = swap_function[3];
               break;
             }
             /*}}}*/
      case 5:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE5_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE5_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE5_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE5_3) = swap_function[3];
               break;
             }
             /*}}}*/
      case 6:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE6_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE6_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE6_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE6_3) = swap_function[3];
               break;
             }
             /*}}}*/
      case 7:
             /*{{{*/
             {
               *((unsigned int *)MUX_PAGE7_0) = swap_function[0];
               *((unsigned int *)MUX_PAGE7_1) = swap_function[1];
               *((unsigned int *)MUX_PAGE7_2) = swap_function[2];
               *((unsigned int *)MUX_PAGE7_3) = swap_function[3];
               break;
             }
             /*}}}*/
  
      default:{break;}
    }
  /*}}}*/
}
/*}}}*/

void set_mux_0_transparent(void)
/*{{{*/
{
  *((unsigned int *)MUX_PAGE0_0)=0x40C80000; /* sets transparent mux */
  *((unsigned int *)MUX_PAGE0_1)=0x51D90000;
  *((unsigned int *)MUX_PAGE0_2)=0x62EA0000;
  *((unsigned int *)MUX_PAGE0_3)=0x73FB0000;
}
/*}}}*/

void program_mux_map_ram(int which_swap,void *address)
/*{{{*/
{
int map_entry;
  which_swap = (which_swap & 0x07) << 4;
  map_entry = shadow_read_map_ram(address);
  map_entry = (map_entry & 0xFF80) | which_swap;
  shadow_write_map_ram(address,map_entry);
}
/*}}}*/
