/*{{{ COMMENT Standard Header*/
/******************************************************************
COPYRIGHT (C) INMOS LIMITED - 1990, 1991

    Source file name : rom.c
              Author : S.R.Christian, D. Johnston
   Date first issued : 12-Apr-91

Routines to support Flash ROM devices on IMS B016

Date        Modification                                    Initials
----        ------------                                    --------
02-Oct-91   Adapted from original                           SRC
06-Oct-91   Delay process now reads required delay
            Other erase, programming functions tidied       SRC
******************************************************************/
/*}}}*/
#include "basic.h"
#include "rom.h"
#include <channel.h>
#include <process.h>

static int delay_started=0;
Channel a;
Channel b;
Channel *delaychan=&a;
Channel *delayack=&b;

void delayproc(Process *p)
/*{{{*/
{
  int delay;
  for(;;) {
    delay = ChanInInt(delaychan);
    ProcWait(delay);
    ChanOutInt(delayack,0);
  }
}
/*}}}*/
void delay_us(int delay)
/*{{{*/
{
int dummy;
Process *p;
  if (delay_started==0)
    {
      ChanInit(delaychan);
      ChanInit(delayack);
      p=ProcAlloc(delayproc,100,0);
      ProcRunHigh(p);
      delay_started=1;
    }
  ChanOutInt(delaychan,delay); 
  dummy=ChanInInt(delayack);   

}
/*}}}*/
int blank_check_rom(void)
/*{{{*/
{
/* checks that ROM contents are uniformly erased, or returns address in error */
int            i;
unsigned long  contents, first_contents;
int            size;
long*          address;

  turnon_vpp();
  size = find_rom_size();
  reset_rom();
  /* work from bottom of ROM decode space */
  address = (long*)ROM_BASE_ADDRESS;
  first_contents = *address;
  for(i=0;i < (size/4);i++) {
    contents =*address;
    if (contents != first_contents) 
      return((int)address);
    address++;
  }
  /* return byte for erased or zeroed */
  if (first_contents == 0xffffffff)
    return(0xff);
  if (first_contents == 0x0)
    return(0);
  else
    return(-1);
}
/*}}}*/
void reset_rom(void)
/*{{{*/
{
  turnon_vpp();
  write_rom(0,0,0xFF);
  write_rom(0,0,0xFF);
  write_rom(1,0,0xFF);
  write_rom(1,0,0xFF);
  write_rom(2,0,0xFF);
  write_rom(2,0,0xFF);
  write_rom(3,0,0xFF);
  write_rom(3,0,0xFF);
  rom_read_mode(0);
  rom_read_mode(1);
  rom_read_mode(2);
  rom_read_mode(3);
  turnoff_vpp();
}
/*}}}*/
void write_rom(int which_rom,int address,int data)
/*{{{*/
{
  *((unsigned int *)(ROM_BASE_ADDRESS + address))
    = ((data & 0xff) << (8 * which_rom));

}
/*}}}*/
int read_rom(int which_rom,int address)
/*{{{*/
{
int i;
  i=*((unsigned int *)(ROM_BASE_ADDRESS+address));
  i=(i >> (8 * which_rom));
  i=i & 0xFF;
  return(i);
}
/*}}}*/
void turnon_vpp(void)
/*{{{*/
{
  *((unsigned int *)ENABLE_VPP)=0x100; /* turn on the vpp generator */
  ProcWait(1600);                     /* wait 100 msec */
}
/*}}}*/
void turnoff_vpp(void)
/*{{{*/
{
  *((unsigned int *)ENABLE_VPP)=0x0000; /* turn on the vpp generator */
  ProcWait(100);                     /* wait 100 msec */
}
/*}}}*/
int read_rom_manf_code(int which_rom)
/*{{{*/
{
int dev;
          write_rom(which_rom,0,READ_MANF_CODE);
          dev=read_rom(which_rom,0);
          return(dev);
}
/*}}}*/
int read_rom_device_code(int which_rom)
/*{{{*/
{
int dev;
          write_rom(which_rom,0,READ_MANF_CODE);
          dev=read_rom(which_rom,4);
          return(dev);
}
/*}}}*/
int find_rom_size(void)
/*{{{*/
{
int dev;
  dev=read_rom_device_code(0); /* arbitrary rom number 0-3 */
  if (dev==0xb2) return(0x20000);  /* Intel 256K bit */
  if (dev==0xb4) return(0x80000);  /* Intel 1M bit */
  if (dev==0xb8) return(0x40000);  /* Intel 512K bit */
  if (dev==0xa1) return(0x20000);  /* AMD 256K bit */
  if (dev==0xa7) return(0x80000);  /* AMD 1M bit */
  if (dev==0x25) return(0x40000);  /* AMD 512K bit */
  if (dev==0x2a) return(0x100000); /* AMD 2M bit */
  else         return(-1);
}
/*}}}*/
void rom_read_mode(int which_rom)
/*{{{*/
{
          write_rom(which_rom,0,0);
}
/*}}}*/
int program_wait_verify(int which_rom,int address,int data)
/*{{{*/
{
int temp;
          write_rom(which_rom,0,SETUP_PROG);
          write_rom(which_rom,address,data);

          delay_us(100);
          write_rom(which_rom,address,PROG_VERIFY);
          delay_us(6);
          temp=read_rom(which_rom,address);
          if (temp != data)
                    return(-1);
          else
                    return (0);
}
/*}}}*/
int program_byte(int which_rom,int address,int data)
/*{{{*/
{
  int x;
  int word_addr;

  /* make sure we address the correct bytes */
  which_rom = which_rom & 0x3;
  rom_read_mode(which_rom);
  word_addr = address & 0xfffffffc;
  if (read_rom(which_rom,word_addr) == data)
     return(0);

  for (x = 0 ; x < 25; x++) {
    if (program_wait_verify(which_rom, word_addr, data) == 0)
      return(0);
  }
  return (-1);
}
/*}}}*/
void erase_wait(int which_rom)
/*{{{*/
{
  /* write erase code to ROM and wait for result */
  ProcWait(10);
  write_rom(which_rom,0,SETUP_ERASE);
  write_rom(which_rom,0,ERASE);
  ProcWait(160);                    /* Wait 10ms */
}
/*}}}*/
int erase_verify(int which_rom,int address)
/*{{{*/
{
  write_rom(which_rom,address,ERASE_VERIFY);
  delay_us(6);
  if (read_rom(which_rom,address) != 0xff)
     return(-1);
  else
     return (0);
}
/*}}}*/
int erase_rom(int which_rom)
/*{{{*/
{
  /* Vpp should be on when this routine is called */
  int size,address,pulse_count,do_pulse;

  rom_read_mode(which_rom);
  size = find_rom_size();
  /* program all locations to zero at first */
  for (address = 0; address < size; address+=4) {
    if (program_byte(which_rom,address,0) == -1) 
      return(-1);
  }
  /* now force erase cycles */
  pulse_count = 0;
  do_pulse = 1;
  for (address = 0; address < size; ) {
    if (do_pulse == 1) {
      erase_wait(which_rom);
      pulse_count++;
      do_pulse = 0;
    }
    if (erase_verify(which_rom,address) == -1)
      do_pulse = 1;
    else
      address+=4;
    if (pulse_count > 1000)
      break;
  }
  if (address < size)
    return(-1);
  else
    return (0);
}
/*}}}*/
