/*{{{ COMMENT Standard Header*/
/******************************************************************
COPYRIGHT (C) INMOS LIMITED - 1990, 1991

    Source file name : rtc.c
              Author : S.R.Christian, D. Johnston
   Date first issued : 12-Apr-91



Date        Modification                                    Initials
----        ------------                                    --------
02-Oct-91   Adapted from original                           SRC
25-Oct-91   Added protection for RTC funny values           SRC
******************************************************************/
/*{{{Info*/
/* 
   These routines are designed to work with the time routines
   in the C toolset library, defined in <time.h>.

   #include <time.h>

   rtc_settimeofday(time_t *time);

   rtc_gettimeofday(time_t *time);

   rtc_settimeofday sets the current time, expressed as seconds since
   midnight on January 1 1970. It returns -1 if the time can not be
   set. rtc_gettimeofday returns the current time, or -1 if the time
   can not be read.


*/
/*}}}*/

/*}}}*/

#include <time.h>
#include <string.h>
#include "rtc.h"
#include "basic.h"

#define isleap(y) (((y) % 4) == 0 && ((y) % 100) != 0 || ((y) % 400) == 0)

#include <stdio.h>

/*{{{read and write ops*/
int readshift(int *address)
{
   return(((*address>>8) & 0xFF));
}

void writeshift(int *address,int value)
{
  *address = (value <<8);
}

void writebinary(int *address,int bin_value)
{
  int bcd_value;

  bcd_value = ((bin_value / 10) << 4) & 0xF0;
  bcd_value |= ((bin_value % 10) & 0x0F);

  writeshift(address,bcd_value);
}

int readbinary(int *address)
{
  int temp,x;

  temp = readshift(address);
  x = ((temp >> 4) * 10) + (temp & 0x0F);
  return(x);
}
/*}}}*/

int rtc_get(struct tm *broken_time)
/*{{{*/
{
  int temp;

  if (check_board_init() == 0)
    return(-1);

  /* page 0, reg 0 */
  writeshift((int *)RTC_MAIN_STATUS,(int)0x00);

  writeshift((int *)RTC_TIME_SAVE_CONTROL,(int)0x00);


  /* page 0, reg 1 */
  writeshift((int *)RTC_MAIN_STATUS,(int)0x40);

  /* stop the clock */

  temp = readshift((int *)RTC_REAL_TIME_MODE);
  writeshift((int*)RTC_REAL_TIME_MODE,(int)(temp & 0xF7));

  /* get the year */
  broken_time->tm_year = readbinary((int *)RTC_YEARS_COUNTER);

  /* Start the clock again */
  writeshift((int *)RTC_REAL_TIME_MODE,(int)(temp | 0x08));

  /* page 0, reg 0 */
  writeshift((int *)RTC_MAIN_STATUS,(int)0x00);

  /* get the time into the save RAM */

  writeshift((int *)RTC_TIME_SAVE_CONTROL,(int)0x00);


  broken_time->tm_sec = readbinary((int *)RTC_SECONDS_TIME_SAVE);
  broken_time->tm_min = readbinary((int *)RTC_MINUTES_TIME_SAVE);
  broken_time->tm_hour = readbinary((int *)RTC_HOURS_TIME_SAVE);
  broken_time->tm_mday = readbinary((int *)RTC_DAY_OF_MONTH_TIME_SAVE);
  broken_time->tm_mon = readbinary((int *)(RTC_MONTHS_TIME_SAVE))-1;

  writeshift((int *)RTC_TIME_SAVE_CONTROL,(int)0x80);

  /* detect funny values from RTC  */
  if ((broken_time->tm_year < 70) ||
      (broken_time->tm_mon < 0) ||
      (broken_time->tm_mday < 0) ||
      (broken_time->tm_hour < 0) ||
      (broken_time->tm_min < 0) ||
      (broken_time->tm_sec < 0))
    return(-1);

  return(0);
}
/*}}}*/

int rtc_set(struct tm *broken_time)
/*{{{*/
{
  int x;

  if (check_board_init()==0)
    return(-1);
 
  /* page 0, reg 1 */
  writeshift((int *)RTC_MAIN_STATUS,(int)0x40);

  writeshift((int *)RTC_REAL_TIME_MODE,(int)0x00); /* Stops the clock */
  writeshift((int *)RTC_OUTPUT_MODE,(int)0x00);
  writeshift((int *)RTC_INTERRUPT_CONTROL_0,(int)0x00);
  writeshift((int *)RTC_INTERRUPT_CONTROL_1,(int)0x00);

  writebinary((int *)RTC_100_SECOND_COUNTER,0);
  writebinary((int *)RTC_SECONDS_COUNTER,broken_time->tm_sec);
  writebinary((int *)RTC_MINUTES_COUNTER,broken_time->tm_min);
  writebinary((int *)RTC_HOURS_COUNTER,broken_time->tm_hour);
  writebinary((int *)RTC_DAY_OF_MONTH_COUNTER,broken_time->tm_mday);
  writebinary((int *)RTC_DAY_OF_WEEK_COUNTER,broken_time->tm_wday + 1);
  writebinary((int *)RTC_MONTHS_COUNTER,broken_time->tm_mon + 1);
  writebinary((int *)RTC_YEARS_COUNTER,broken_time->tm_year);

  /* Clear out the save registers to remove corrupt data */

  writebinary((int *)RTC_SECONDS_TIME_SAVE,0);
  writebinary((int *)RTC_MINUTES_TIME_SAVE,0);
  writebinary((int *)RTC_HOURS_TIME_SAVE,0);
  writebinary((int *)RTC_DAY_OF_MONTH_TIME_SAVE,0);
  writebinary((int *)RTC_MONTHS_TIME_SAVE,0);

  /* Find out how long since a leap year */

  for (x=0;x <=3;x++)
    if (isleap(broken_time->tm_year - x))
      break;

  /* Start the clock going */
  writeshift((int *)RTC_REAL_TIME_MODE,(int)(0x08 | x));
  return(0);
}
/*}}}*/

static period_set=FALSE; /* Test flag to check if period reg has been set */

time_t rtc_settimeofday(time_t* time)
/*{{{*/
{
  struct tm *broken_time;

  if (period_set==FALSE) {
    period_set=TRUE;
    writeshift((int *)RTC_PERIODIC_FLAG,(int)0x00);
  }
  broken_time = localtime(time);
  if (rtc_set(broken_time) == -1)
    return((time_t)-1);
  return(*time);
}
/*}}}*/

time_t rtc_gettimeofday(time_t* time)
/*{{{*/
{
  struct tm broken_time;

  memset(&broken_time, 0, sizeof(broken_time));

  if (rtc_get(&broken_time) == -1){
    *time = (time_t)-1;
    return(*time);
  }
  *time = mktime(&broken_time);
  return (*time);
}
/*}}}*/
