/*{{{ COMMENT Standard Header*/
/******************************************************************
COPYRIGHT (C) INMOS LIMITED - 1990, 1991

    Source file name : boardmon.c
              Author : S.R.Christian, D. Johnston
   Date first issued : 2-Aug-90

Contains main routines for a PROM based monitor using the on-board duart

Date        Modification                                    Initials
----        ------------                                    --------
02-Oct-91   Adapted from original monitor.c by DJ           SRC
******************************************************************/
/*}}}*/

/* NB use reduced library for PROM based operation */
#include <stdiored.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <iocntrl.h>
#include <misc.h>
#include <process.h>
#include <time.h>
#include "vme.h"
#include "duart.h"
#include "event.h"
#include "basic.h"
#include "rtc.h"
#include "rom.h"
#include "clilib.h"

/*{{{static variables*/
static char last_command[max_tok_len];
static berr_flag=FALSE;


/*}}}*/

/*{{{console io routines for DUART A*/
int io_write(char* buffer)
/* support for 'console' string output using duart port A */
{
   duart_write_a(buffer);
   return(0);
}

int io_read(char* prompt, char* buffer,long buflen)
/* support for a read of a null terminated input line from duart port A */
{
  char      c;
  short     cnt;
  boolean   eol;

  io_write(prompt); 
  
  cnt = 0;
  eol = false;
  buffer[0] = '\0';
  c=' ';

  /* read characters one by one and perform line editing functions before
     putting them in the read buffer. Characters are echoed as accepted. */
  while((c != '\0') && !eol && (cnt < buflen)) {
    c = duart_getchar_a();
    if (isprint(c)) {
      buffer[cnt] = c;
      cnt++;
      duart_putchar_a(c);
    }
    else switch(c) {
      /* control characters are recognised and dealt with. Escape
         is used to repeat the previous line of input */
      case '\r' :
      case '\n' : buffer[cnt++] = '\0';
                  duart_putchar_a('\r');
                  duart_putchar_a('\n');
                  eol = true;
                  strcpy(last_command, buffer);
                  break;
      case '\b' : if (cnt > 0) {
                    cnt--;
                    duart_putchar_a('\b');
                    duart_putchar_a(' ');
                    duart_putchar_a('\b');
                  }
                  break;
      case 0x1b : strcpy(buffer, last_command);
                  duart_putchar_a('\r');
                  duart_putchar_a('\n');
                  duart_write_a(buffer);
                  cnt = strlen(buffer);
                  break;
    }
  }

  return(cnt);
}
/*}}}*/

/*{{{void io_setup(void);*/
void io_setup(void)
  {

    duart_setup(DUART_RX_BAUD_B,BAUD_9600);
    duart_setup(DUART_TX_BAUD_B,BAUD_9600);
    duart_setup(DUART_RESET_ERROR_B,0);
    duart_setup(DUART_ENABLE_TX_B,FALSE);
    duart_setup(DUART_ENABLE_RX_B,FALSE); /* RESET ERROR, DISABLE RX & TX */
    duart_setup(DUART_RESET_TX_B,0);   /* RESET tX */
    duart_setup(DUART_RESET_RX_B,0);   /* RESET rX */
    duart_setup(DUART_DATA_BITS_B,8);
    duart_setup(DUART_STOP_BITS_B,1);
    duart_setup(DUART_PARITY_B,DUART_NONE);
    duart_setup(DUART_ENABLE_TX_B,TRUE);
    duart_setup(DUART_ENABLE_RX_B,TRUE); /* ENABLE RX & TX */

    duart_setup(DUART_RX_BAUD_A,BAUD_9600);
    duart_setup(DUART_TX_BAUD_A,BAUD_9600);
    duart_setup(DUART_RESET_ERROR_A,0);
    duart_setup(DUART_ENABLE_TX_A,FALSE);
    duart_setup(DUART_ENABLE_RX_A,FALSE); /* RESET ERROR, DISABLE RX & TX */
    duart_setup(DUART_RESET_TX_A,0);      /* RESET tX */
    duart_setup(DUART_RESET_RX_A,0);      /* RESET rX */
    duart_setup(DUART_DATA_BITS_A,8);
    duart_setup(DUART_STOP_BITS_A,1);
    duart_setup(DUART_PARITY_A,DUART_NONE);
    duart_setup(DUART_ENABLE_TX_A,TRUE);
    duart_setup(DUART_ENABLE_RX_A,TRUE); /* ENABLE RX & TX */

  }
/*}}}*/
/*{{{void my_berr_handler(void);*/
void my_berr_handler(void)
{
    /* *(unsigned int *)SPLAT_BERR=0x01; old code in error? */
    *(unsigned int *)SPLAT_BERR=0x100;
    berr_flag=TRUE;
}
/*}}}*/

#include "moncoms.c"

/*{{{command setup*/
static boolean command_setup()
/* setup routine for IMS B016 monitor commands */
{

  register_command("DISPLAY", display_memory, 
                   "<address><range> Displays memory contents");
  register_command("FIND", find_memory, 
                   "<address><range><value> Searchs for memory contents");
  register_command("FILL", fill_memory, 
                   "<address><range><value> Fill memory with value");
  register_command("COPY", copy_memory, 
                   "<address><range><newaddress> Copies memory contents");
  register_command("MAPRAM", mapram, 
                   "<address> Display of mapram entry");
  register_command("MUXRAM", set_mux, 
                   "<address><range><mapcode><rswop><wswop> Set up mux & mapram entry");
  register_command("SLAVE", setup_slave, 
                   "<address><range><addwidth> Set up slave access region");
  register_command("ROM", read_rom_status, 
                   "Display of ROM status");
  register_command("BWRITE", write_duart, 
                   "<string> Writes to duart port B");
  register_command("BREAD",  read_duart, 
                   "Reads string from duart port B");
  register_command("RTC",  read_rtc, 
                   "<timeval> Sets and/or reads RTC registers");
  register_command("INIT",  reset_ss, 
                   "<analflag><linkid> Reset/analyse on subsystem port");
  register_command("PEEK",  peek_ss, 
                   "<address><range> Displays remote memory");
  register_command("POKE",  poke_ss, 
                   "<address><range><value> Sets remote memory");

  /* utility default variables */
  assign_integer("MEMSIZE", 0x400000, true);
  assign_integer("BASEADDRESS", 0x80000000, false);
  assign_integer("RANGE", 0x100, false);
  assign_integer("DATAVALUE", 0, false);
  assign_integer("ADDRESSVALUE", 0x80000000, false);

  /* H/W predefines for mapram */

  assign_integer("MASTER", MASTER_PERMIT, true);
  assign_integer("D8BIT", DSIZ_8BIT, true);
  assign_integer("D16BIT", DSIZ_16BIT, true);
  assign_integer("D32BIT", DSIZ_32BIT, true);
  assign_integer("A32BIT", ASIZ_32BIT, true);
  assign_integer("A24BIT", ASIZ_24BIT, true);
  assign_integer("A16BIT", ASIZ_16BIT, true);
  assign_integer("LADD00", LADD00, true);
  assign_integer("LADD01", LADD01, true);
  assign_integer("LADD10", LADD10, true);
  assign_integer("LADD11", LADD11, true);

  /* MUX functions */
  assign_integer("FN1", 0x10, true);
  assign_integer("FN2", 0x20, true);
  assign_integer("FN3", 0x30, true);
  assign_integer("FN4", 0x40, true);
  assign_integer("FN5", 0x50, true);
  assign_integer("FN6", 0x50, true);
  assign_integer("FN7", 0x70, true);

  return(false);
}
/*}}}*/
/*{{{main*/
int main()
{

   /* do board level initialisation */

   basic_board_init();
   start_event_handler();
   install_berr_handler(my_berr_handler);
   
   /* initialise on board io */
   duart_init();  
   io_setup();    
   duart_use_xon_a(TRUE);
   duart_use_xon_b(TRUE);

   /* assert master role for case where we are plugged into slot 1 of bus
      and jumper is fitted */
   init_slot_one(ARBITER_PRI,VMETIMEOUT_64US);
   init_vme_master(3, RELMODE_BCAP, false);
   set_mux_0_transparent();


   /* initialise buffer for command editing support */
   strcpy(last_command,"help");

   /* set up never ending monitor */
   while (true) {
      
     /* print title and start interactive loop */
     print("\nIMS B016 board monitor v1.0\n\n");
     cli_main(command_setup, 1000, 16, "B016> ", "");
   }  
}
/*}}}*/

