/*{{{ COMMENT Standard Header*/
/******************************************************************
COPYRIGHT (C) INMOS LIMITED - 1990, 1991

    Source file name : vrtxmon.c
              Author : S.R.Christian, D. Johnston
   Date first issued : 2-Aug-90

Contains main routines for a monitor cli running under VRTX32/T
and using the IMS B016 DUART facilities.

Date        Modification                                    Initials
----        ------------                                    --------
02-Oct-91   Adapted from original monitor.c by DJ           SRC
14-Oct-91   Added VRTX specific support structure           SRC
******************************************************************/
/*}}}*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <iocntrl.h>
#include <process.h>
#include <channel.h>
#include <misc.h>
#include <time.h>
#include "vme.h"
#include "duart.h"
#include "event.h"
#include "basic.h"
#include "rtc.h"
#include "rom.h"
#include "clilib.h"

#include "vrtxvisi.h"

#define RTSCOPE_HOTKEY 0x1  /* Control A defined to start RTSCOPE */

/*{{{static variables*/
static char last_command[max_tok_len];
static berr_flag=FALSE;


/*}}}*/

/*{{{void my_berr_handler*/
void my_berr_handler(void)
{
    /* *(unsigned int *)SPLAT_BERR=0x01; old code in error? */
    *(unsigned int *)SPLAT_BERR=0x100;
    berr_flag=TRUE;
}
/*}}}*/

/*{{{break_handler_low*/
void break_handler_low()
/* preface to VRTX debugger - must be in user code space since it uses a
   static array as a stack, and this has to be initialised by the user 

   The first time this code is entered is via a gratuitous j 0 instruction
   in the init_breakpoint routine. It returns to this code via the next j 0
   inside the main loop of this code. Thereafter, when hitting a breakpoint
   the debugger body code is executed - RTScope - with the appropriate
   register context derived from the break point state. */
{
  int  ABC_Reg[3];
  int  Low_Areg,Low_Breg,Low_Creg;
      
  while (true) {


    /* Load previous registers values here in order to restore state */

    Low_Areg = ABC_Reg[0];
    Low_Breg = ABC_Reg[1];
    Low_Creg = ABC_Reg[2];
    __asm {ldl Low_Creg; ldl Low_Breg; ldl Low_Areg; 
           timerenablel;timerenableh; }

    /* return to user space just like we had never left it */
    __asm { j 0 ;}

    /* Save new register values here, as we enter from a breakpoint */

    __asm {timerdisableh; timerdisablel; stl Low_Areg; 
          stl Low_Breg; stl Low_Creg; }
    ABC_Reg[0] = Low_Areg;
    ABC_Reg[1] = Low_Breg;
    ABC_Reg[2] = Low_Creg;

    /* invoke real debugger here, passing in register values */

    body(&ABC_Reg,1);

  }  
}
/*}}}*/
/*{{{init_break_handler*/

#define IMS_RETURN_WPTR(task_state_p) \
{extern volatile const void *_params;\
 task_state_p= (int*) _params;}

#define BREAK_STACK_SIZE 1024
static int stack[BREAK_STACK_SIZE];  /* space for break handler stack */

void init_break_handler()
/* this is a bit of magic to set up the locations at start of 
   transputer memory which are used by the break point process 
   A stack is initialised to contain a pointer to the static base
   normally passed invisibly between procedures as wptr+1 */
{ 
  int   temp;
  int*  mem_start;
  int*  wptr;

  __asm {ldmemstartval; stl mem_start; }
  IMS_RETURN_WPTR(wptr);
  temp = *(wptr+1);
  /* now we have got hold of the global static base address as passed in this
     function. Now set up the stack area for the breakpoint handler to look
     like it has just been called - i.e. pass it the GSB in the same position */
  stack[BREAK_STACK_SIZE - 1]= temp;
  *(mem_start+2)=(int) &stack[BREAK_STACK_SIZE - 2];
  *(mem_start+3)=(int) break_handler_low;
  __asm {setj0break; j 0;}

}
/*}}}*/

/*{{{console_proc*/
#define BUFFER_SIZE 120
/* static channel definition for console process communication */
static  Channel *rx_chr;
static  Channel *tx_chr;
static  Channel *rq_chr;
/* connections to host from config file */
static  Channel *c_input;
static  Channel *c_output;

/*{{{rtscope_proc*/
static Channel *rts; /* channel to toggle this process */

void rtscope_proc()
/* run RTSCOPE from async process that locks scheduler. console_proc
   then runs independently to service RTSCOPE requests */
{
  while(true) {
    ChanInChar(rts);
    ui_enter();
    DB_GO();
    ui_exit();
    ProcWait(100000);
  }
}
/*}}}*/

void console_proc(Process *p)
/* Console (i.e. host) input & output buffering process. Set up at high priority
   to alleviate inefficiencies in single character RTScope console I/O */
{
  int     time;
  short   in_len;
  short   out_len;
  char    message[BUFFER_SIZE+8];  /* 7 + BUFFER_SIZE is largest : RGI */     
  char    buffer[20];
  short   cnt;
  char    c;
  int     wait_for_host, message_full;
  int     read_pending, read_action ;
  int     write_pending, write_action ;
  int     done_keypoll;
  int     res;
  Process *r = NULL;

 
  /* set up write operation message buffer */
  message[0]=(char)24;  /* FputBlock function */
  message[1]=(char)1;   /* stdout stream */
  message[2]=(char)0;
  message[3]=(char)0;
  message[4]=(char)0;
  message[5]=(char)0;   /* length byte */
  message[6]=(char)0;
  message[7]=(char)0;   /* data start */

  cnt = 0;
  out_len = 0;
  done_keypoll = false;
  wait_for_host = false;
  read_pending = false;
  write_pending = false;
  message_full = false;
  read_action = false;
  write_action = false;

  while(true) {
    /* all may seem complex, but loop works as a state machine for
       both input and output requests.*/

    /* get current time, calculate timeout and set up alternation */
    /* If the output buffer is full we cannot allow further chars into the
                                buffer otherwise we kill ourselves */
    time= ProcTime();
    if ( message_full )
      res = ProcTimerAlt(ProcTimePlus(time, 100000), 
                       c_input, rq_chr, NULL);
    else
      res = ProcTimerAlt(ProcTimePlus(time, 100000), 
                       c_input, rq_chr, tx_chr, NULL);

    switch(res) {
      case 0: /* return message from host */
              /*{{{deal with return message from request*/
              ChanIn(c_input, &in_len, 2);
              ChanIn(c_input, buffer, in_len);
              if (read_action) {
                /* return from getkey */
                c = buffer[1];
                ChanOutChar(rx_chr, c);
                wait_for_host = false;
                read_action = false ;
              }
              else
              if (write_action) {
                /* return from a write request */
                wait_for_host = false;
                write_action = false;
              }
              else __asm { seterr; }
              /*}}}*/
              break;

      case 2: /* new character from application. update counts and
                 try to issue a write if getting large or special char  */
              /*{{{*/
              c = ChanInChar(tx_chr);
              cnt++;
              message[cnt+6] = c;
              if ((cnt >= BUFFER_SIZE) || (c < ' ')) {
                message_full = true;
                write_pending = true;
              }
              /*}}}*/
              break;

      case 1: /* request from application for character. try to
                 flush output if it exists, then issue a read */
              /*{{{*/
              c = ChanInChar(rq_chr);
              if (cnt > 0)
                write_pending = true;
              read_pending= true;
              /*}}}*/
              break;

      default:/* timeout - write existing buffer block to host, if not empty 
                 else, if idle,  poll for a special key */
              /*{{{*/
              if (cnt > 0)
                write_pending = true; 
              else if (!wait_for_host) {
                in_len = 6 ;
              
                buffer[0]=(char)31;  /* Pollkey - does not suspend if no input */
                buffer[1]=(char)0;
                buffer[2]=(char)0;
                buffer[3]=(char)0;
                buffer[4]=(char)0;
                buffer[5]=(char)0;
              
                ChanOut(c_output, &in_len, 2 );
                ChanOut(c_output, buffer, in_len);
              
                ChanIn(c_input, &in_len, 2 );
                ChanIn(c_input, buffer, in_len );
              
                if ( buffer[0] == 0 ) {
                  c = buffer[1];
                  if (c == RTSCOPE_HOTKEY) {
                    /* if first time, allocate and run process to start RTSCOPE */
                    if (r == NULL) {
                      rts = ChanAlloc();
                      r = ProcAlloc(rtscope_proc, 2000, 0);
                      ProcRunHigh(r);
                    }
                    /* send dummy character to process */
                    ChanOutChar(rts, 0);
                  }
                }
              }
              
              /*}}}*/
              break;
    }

    /*{{{issue pending read and writes*/
    /* now, if we are not in the middle of something, issue read and write
       requests that are pending. Once issued we can accept Tx chars again */
    if (write_pending && !wait_for_host) {
      message[5] = (char)cnt;
      out_len = cnt+7;
      if ((out_len & 0x1) != 0) /* length must be even no */
        out_len++;
      ChanOut(c_output, &out_len, 2);
      ChanOut(c_output, message, out_len);
      cnt = 0;
      message_full = false;
      write_pending = false;
      write_action = true ;
      wait_for_host = true;
    }
    if (read_pending && !wait_for_host) {
      in_len = 6;
      buffer[0]=(char)30;  /* Getkey request */
      buffer[1]=(char)0;
      buffer[2]=(char)0;
      buffer[3]=(char)0;
      buffer[4]=(char)0;
      buffer[5]=(char)0;
      ChanOut(c_output,&in_len,2);
      ChanOut(c_output,buffer,in_len);
      read_pending = false;
      read_action = true ;
      wait_for_host = true;
    }
    /*}}}*/
  }
}
/*}}}*/
/*{{{console I/O interface*/
long console_input()
/* interface to RTScope character input */
{
  char c;

  ChanOutChar(rq_chr, 0);
  c = ChanInChar(rx_chr);
  return((long)c);
}


void console_output(long c)
/* interface to RTScope character output */
{
  ChanOutChar(tx_chr, (char)c);
}
/*}}}*/

/*{{{serial_setup*/
void serial_setup(void)
/* disables channel B of DUART and sets A up at 9600 baud */
{
    duart_setup(DUART_RX_BAUD_B,BAUD_9600);
    duart_setup(DUART_TX_BAUD_B,BAUD_9600);
    duart_setup(DUART_RESET_ERROR_B,0);
    duart_setup(DUART_ENABLE_TX_B,FALSE);
    duart_setup(DUART_ENABLE_RX_B,FALSE); /* RESET ERROR, DISABLE RX & TX */
    duart_setup(DUART_RESET_TX_B,0);   /* RESET tX */
    duart_setup(DUART_RESET_RX_B,0);   /* RESET rX */

    duart_setup(DUART_RX_BAUD_A,BAUD_9600);
    duart_setup(DUART_TX_BAUD_A,BAUD_9600);
    duart_setup(DUART_RESET_ERROR_A,0);
    duart_setup(DUART_ENABLE_TX_A,FALSE);
    duart_setup(DUART_ENABLE_RX_A,FALSE); /* RESET ERROR, DISABLE RX & TX */
    duart_setup(DUART_RESET_TX_A,0);      /* RESET tX */
    duart_setup(DUART_RESET_RX_A,0);      /* RESET rX */
    duart_setup(DUART_DATA_BITS_A,8);
    duart_setup(DUART_STOP_BITS_A,1);
    duart_setup(DUART_PARITY_A,DUART_NONE);
    duart_setup(DUART_ENABLE_TX_A,TRUE);
    duart_setup(DUART_ENABLE_RX_A,TRUE); /* ENABLE RX & TX */
}
/*}}}*/
/*{{{serial_isr*/
/* static variables used by isr - initialised befor isr started */
static int      duart_int_mask;
static char     rxflag_a;
static char     rxflag_b;


void serial_isr()
/* this VRTX ISR is coded as a function rather than as a process since it is
   called from vectored interrupt system of the IMS B016 VME environment */
{
  int   mask;
  int   isr;
  int   c;
  int   err;

  isr=(*((uint *)DUART_ISR)) & 0xFF;     /* Get interrupt status */
  if (isr != 0) {
    /* Identify what caused the interrupt & wakeup the buffer routines */
    isr=(isr & duart_int_mask);
    if ((isr & 0x01)==0x01) { /* Then transmitter A is ready */

      if (rxflag_a==XOFF) {
        mask=(duart_int_mask & 0x32);
        duart_int_mask=mask;
        *((uint *)DUART_IMR)=mask; /* disable the tx interrupt */
      }
      else {
        /* get character from VRTX32 Xmit buffer via ISR procedure 
           If no character then turn of tx interrupt */
        ui_enter();
        c = ui_txrdy(&err);
        ui_exit();
        if (err == ER_NCP) {
          mask=(duart_int_mask & 0x32);
          duart_int_mask=mask;
          *((uint *)DUART_IMR)=mask; /* disable the tx interrupt */
        }
        else
          *((uint *)DUART_THRA) = c;  /* send the char, clear the tx interrupt */
      }
    }
    if ((isr & 0x10)==0x10) { /* Then transmitter B is ready */

      if (rxflag_b==XOFF) {
        mask=(duart_int_mask & 0x23);
        duart_int_mask=mask;
        *((uint *)DUART_IMR)=mask; /* disable the tx interrupt */
      }
      else {
        ;  /* do nothing */
      }
    }
    if ((isr & 0x02)==0x02) {  /* Then char recieved on port A */

      c=(*((uint *)DUART_RHRA)) & 0xFF;     /* get the char, clear the int */
      if ((c==(char)XON) || (c==(char)XOFF)) {
        rxflag_a=(c & 0x7F);
        if (c==(char)XON) {
          mask=(duart_int_mask | 0x01);
          duart_int_mask=mask;
          *((uint *)DUART_IMR)=mask; /* enable the tx interrupt */
        }
        else {
          mask=(duart_int_mask & 0x32);
          duart_int_mask=mask;
          *((uint *)DUART_IMR)=mask; /* disable the tx interrupt */
        }
      }
      else {
        /* send character to VRTX32 via ISR mechanism if its not special */
        ui_enter();
        if (c == RTSCOPE_HOTKEY)
          DB_GO();
        else
          ui_rxchr(c, &err);
        ui_exit();
      }
    }
    if ((isr & 0x20)==0x20) {  /* Then char recieved on port B */

      c=(*((uint *)DUART_RHRB)) & 0xFF;     /* get the char, clear the int */
      if ((c==(char)XON) || (c==(char)XOFF)) {
        rxflag_b=(c & 0x7F);
        if (c==(char)XON) {
          mask=(duart_int_mask | 0x10);
          duart_int_mask=mask;
          *((uint *)DUART_IMR)=mask; /* enable the tx interrupt */
        }
        else {
          mask=(duart_int_mask & 0x23);  
          duart_int_mask=mask;
          *((uint *)DUART_IMR)=mask; /* disable the tx interrupt */
        }
      }
      else {
        ;   /* do nothing */
      }
    }
  }
}   

/*}}}*/
/*{{{TXRDY*/
void TXRDY(int c)
/* this routine is called as the first character is put in a VRTX
   TX buffer and is intended to start up a sequence of interrupt
   controlled I/O. */
{
   *((uint *)DUART_THRA) = c;  /* send the char */
   duart_int_mask = (duart_int_mask | 0x01);
   *((uint*)DUART_IMR)=duart_int_mask; /* enable the tx interrupt */

}
/*}}}*/
/*{{{init_serial*/
void init_serial()
/* initialises on board UART and sets up interrupt handler and
   character buffer processes */
{
    duart_int_mask=0x00;
    rxflag_a = 0;
    rxflag_b = 0;

    *((uint *)DUART_IMR)=duart_int_mask;  /* Interrupts off */

    *((uint *)DUART_CRA)=0x20; /* reset reciever a */
    *((uint *)DUART_CRA)=0x30; /* reset transmitter a */
    *((uint *)DUART_CRA)=0x40; /* reset error status a*/
    *((uint *)DUART_CRA)=0x50; /* reset chan a break change */

    *((uint *)DUART_CRB)=0x20; /* reset reciever b */
    *((uint *)DUART_CRB)=0x30; /* reset transmitter b */
    *((uint *)DUART_CRB)=0x40; /* reset error status b */
    *((uint *)DUART_CRB)=0x50; /* reset chan b break change */

    *((uint *)DUART_CRA) = 0x10;     /* point at m1ra */
    (*((uint *)DUART_MR1A))=0x12;    /* set RxRDY rather than FFULL */

    *((uint *)DUART_CRB) = 0x10;     /* point at m1ra */
    (*((uint *)DUART_MR1B))=0x12;    /* set RxRDY rather than FFULL */

    /* install isr in vectored interrupt system */
    install_duart_int_handler(serial_isr);

    duart_int_mask=0x33;  /* enable input & output interrupts */
    *((uint *)DUART_IMR)=duart_int_mask;

}
/*}}}*/
/*{{{timer_isr*/
void timer_isr(Process *p)
/* generic VRTX32/T ISR routine - high priority process
   waiting on transputer timer. Ticks VRTX clock every 10ms */ 
{
  int  last_time;

  /* keep track of last time to make tick wait time accurate and constant */
  last_time = ProcTime();
  while (true) {
    /* use a routine to do the addition to cope with time value rollover */
    last_time = ProcTimePlus(last_time, 10000);
    ProcAfter(last_time);
    ui_enter();
    ui_timer();
    ui_exit();
  }
}
/*}}}*/
/*{{{config_vrtx*/

CFTBL vrtx_cfg; /* static table space for configuration data */
CFTBL *cfg; /* pointer to satisfy undefine external in library!! */

static struct  DB_CONFIG {
  char*    DbWs;
  long     WS_size;
  char*    VrtxWS;
  long     page_size;
} dbcnf;

extern char *W_ptr; 

void config_vrtx()
/* set up VRTX32/T configuration for monitor application */
{
  long    start_ws;
  int     err;

  /* Fill config table with required data values */
  /* This config data should not need altering for many applications */

  cfg = &vrtx_cfg; /* initialise externally referenced pointer */

  vrtx_cfg.cfwssize=0x30000;              /* workspace size           */
  vrtx_cfg.cfutskct=0x10;                 /* max number of tasks        */
  vrtx_cfg.cfustksz=0x2000;               /* task stack size      */
  vrtx_cfg.cfidle=0x100;                  /* idle task stack size          */
  vrtx_cfg.cfrsrvd1=0;                    /* reserved(must be 0)       */
  vrtx_cfg.cfrsrvd2=0;                    /* reserved(must be 0)       */
  vrtx_cfg.cfrsrvd3=0;                    /* reserved(must be 0)       */
  vrtx_cfg.cfrsrvd4=0;                    /* reserved(must be 0)       */
  vrtx_cfg.cfcbcount=0x30;                /* max id in the system       */
  vrtx_cfg.cftcreate=0;                   /* TCREATE routine pointer     */
  vrtx_cfg.cftdelete=0;                   /* TDELETE routine pointer     */
  vrtx_cfg.cftswitch=0;                   /* TSWITCH routine pointer     */
  vrtx_cfg.cftxrdy=TXRDY;                 /* TXRDY driver pointer       */

  /* allocate a workspace for VRTX32/T private use */
  if((start_ws=(long)malloc(vrtx_cfg.cfwssize))== 0) { 
    exit(1);
  }

  vrtx_cfg.cfwsaddr=(char*)start_ws;

  /* initialise VRTX32/T operation */
  vrtx_init(&vrtx_cfg, &err);

  /* initialise RTScope environent */

#define DBWSSIZE 0x9000

  if((W_ptr=(char *)malloc(DBWSSIZE))== NULL) {
    exit(1);
  }

  /* Initialise configuration variables */

  dbcnf.DbWs      = W_ptr;          /* RTscope workspace addr.    */
  dbcnf.WS_size   = DBWSSIZE;       /* RTscope workspace size     */
  dbcnf.VrtxWS    = (char*)start_ws;/* VRTX32/T workspace addr.   */
  dbcnf.page_size = 20;             /* Screen output line count */

  /* Initialise the RTscope/T environment */
  err = DB_INIT(&dbcnf);  
}
/*}}}*/


/*{{{display_memory*/
boolean display_memory(parse_t *pars_p, char *result_sym_p)
/* does an interactive formatted display of memory contents in numeric
   and ascii format */
{
  boolean   error;
  long      base;
  long      length;
  long      def_base;
  long      def_range;
  long      address;

  /* get default values from symbol variables */
  evaluate_integer("BASEADDRESS", &def_base, 10);
  evaluate_integer("RANGE", &def_range, 10);
    
  error = cget_integer(pars_p, def_base, &base);
  if (error) {
    tag_current_line(pars_p, "Invalid base address");
  }
  else {
    error = (cget_integer(pars_p, def_range, &length) || (length <= 0));
    if (error)
      tag_current_line(pars_p, "Invalid length");
    else {
      /* do a formatted display, address and four values a line.
         Watch for a bus error if we stray out of legal memory */
      long cnt;
      char  contents[17];
      int   i;
      
      address = base;
      cnt = 0;
      while ((cnt <= length) && !berr_flag) {
        print("#%8x: %8x %8x %8x %8x", address,
          *((long*)address),*(((long*)address)+1),
          *(((long*)address)+2),*(((long*)address)+3));
        for (i=0;i<16;i++) {
          contents[i] = *(((char*)address)+i);
          if (isprint(contents[i]) == 0)
            contents[i] = '.';
        }
        contents[16] = '\0';
        print(" (%s)\n", contents);
        /* protect ourselves from address rollover */
        if (address >= 0x7ffffff0)
          address = (long)0x80000000;
        else
          address += 16;
        cnt +=16;
      }
      if (berr_flag) {
        print("Bus error occurred on or near address #%8x\n", address);
        berr_flag = false;
        error = true;
      }
    }
  }
  return(error || assign_integer(result_sym_p, address, false));
}
/*}}}*/
/*{{{find_memory*/
boolean find_memory(parse_t *pars_p, char *result_sym_p)
/* does an search for specific memory contents */

{
  boolean   error;
  long      base;
  long      length;
  long      value;
  long      def_base;
  long      def_range;
  long      def_value;

  /* get default values from symbol variables */
  evaluate_integer("BASEADDRESS", &def_base, 10);
  evaluate_integer("RANGE", &def_range, 10);
  evaluate_integer("DATAVALUE", &def_value, 10);
    
  error = cget_integer(pars_p, def_base, &base);
  if (error) 
    tag_current_line(pars_p, "Invalid base address");
  else {
    error = (cget_integer(pars_p, def_range, &length) || (length <= 0));
    if (error)
      tag_current_line(pars_p, "Invalid length");
    else {
      error = cget_integer(pars_p, def_value, &value);
      if (error)
        tag_current_line(pars_p, "Invalid data value");
      else {
        /*{{{search and display*/
        /* do a search and display results if any*/
        long* address;
        int  cnt;
        
        cnt = 0;
        address = (long*) base;
        while (((long)address < (base + length)) && !berr_flag) {
          if (*address == value){
            print("#%8x : #%8x (%ld)\n", (long)address, *address, *address);
            cnt++;
          }
          address++;
        }
        if (berr_flag) {
          print("Bus error occurred on or near address #%8x\n", address);
          berr_flag = false;
          error = true;
        }
        if (cnt > 0)
          print("\n%d matches in %ld words\n", cnt, length/4);
        /*}}}*/
      }
    }
  }
  return(error || assign_integer(result_sym_p, base, false));
}
/*}}}*/
/*{{{fill_memory*/
boolean fill_memory(parse_t *pars_p, char *result_sym_p)
/* writes specific memory contents */

{
  boolean   error;
  long      base;
  long      length;
  long      value;
  long      def_base;
  long      def_range;
  long      def_value;

  /* get default values from symbol variables */
  evaluate_integer("BASEADDRESS", &def_base, 10);
  evaluate_integer("RANGE", &def_range, 10);
  evaluate_integer("DATAVALUE", &def_value, 10);
    
  error = cget_integer(pars_p, def_base, &base);
  if (error) 
    tag_current_line(pars_p, "Invalid base address");
  else {
    error = (cget_integer(pars_p, def_range, &length) || (length <= 0));
    if (error)
      tag_current_line(pars_p, "Invalid length");
    else {
      error = cget_integer(pars_p, def_value, &value);
      if (error)
        tag_current_line(pars_p, "Invalid data value");
      else {
        /*{{{fill words*/
        /* do a word oriented global fill operation*/
        long* address;
        int  cnt;
        
        cnt = 0;
        address = (long*) base;
        while (((long)address < (base + length)) && !berr_flag) {
          *address = value;
          address++;
        }
        if (berr_flag) {
          print("Bus error occurred on or near address #%8x\n", address);
          berr_flag = false;
          error = true;
        }
        
        /*}}}*/
      }
    }
  }
  return(error || assign_integer(result_sym_p, base, false));
}
/*}}}*/
/*{{{copy_memory*/
boolean copy_memory(parse_t *pars_p, char *result_sym_p)
/* copies memory contents to new address range*/

{
  boolean   error;
  long      base;
  long      length;
  long      newbase;
  long      def_base;
  long      def_range;
  long      def_newbase;

  /* get default values from symbol variables */
  evaluate_integer("BASEADDRESS", &def_base, 10);
  evaluate_integer("RANGE", &def_range, 10);
  evaluate_integer("ADDRESSVALUE", &def_newbase, 10);
    
  error = cget_integer(pars_p, def_base, &base);
  if (error) 
    tag_current_line(pars_p, "Invalid base address");
  else {
    error = (cget_integer(pars_p, def_range, &length) || (length <= 0));
    if (error)
      tag_current_line(pars_p, "Invalid length");
    else {
      error = (cget_integer(pars_p, def_newbase, &newbase) ||
               (newbase == base));
      if (error)
        tag_current_line(pars_p, "Invalid new base address");
      else {
        /*{{{move memory*/
        /* do a memory copy operation using routine that works in overlapped case*/
        
        memmove((char*)newbase, (char*)base, (int) length);
        if (berr_flag) {
          print("Bus error occurred during copy! \n");
          berr_flag = false;
          error = true;
        }
        
        /*}}}*/
      }
    }
  }
  return(error || assign_integer(result_sym_p, base, false));
}
/*}}}*/
/*{{{mapram*/
boolean mapram(parse_t *pars_p, char *result_sym_p)
/* does an formatted display of mapram memory contents in numeric
   and ascii format */
{
  boolean   error;
  long      base;
  long      def_base;
  long      mapram_entry;

  /* get default values from symbol variables */
  evaluate_integer("BASEADDRESS", &def_base, 10);
    
  error = cget_integer(pars_p, def_base, &base);
  if (error) {
    tag_current_line(pars_p, "Invalid base address");
  }
  else {
    /*{{{display data*/
    /* do a formatted display, decoding values as required */
      
    mapram_entry = shadow_read_map_ram((void*)base);
    print("#%8x: %x (", base, mapram_entry);
    if ((mapram_entry & MASTER_PERMIT) > 0) print("MASTER ");
    switch(mapram_entry & 0x6000) {
      case AM_UDATA: print("UDATA "); break;
      case AM_UPROG: print("UPROG "); break;
      case AM_SDATA: print("SDATA "); break;
      case AM_SPROG: print("SPROG "); break;
    }
    switch(mapram_entry & 0x180) {
      case ASIZ_AMSOURCE: print("AMSOURCE "); break;
      case ASIZ_32BIT: print("A32BIT "); break;
      case ASIZ_24BIT: print("A24BIT "); break;
      case ASIZ_16BIT: print("A16BIT "); break;
    }
    switch(mapram_entry & 0x1800) {
      case LADD00: print("LADD00 "); break;
      case LADD01: print("LADD01 "); break;
      case LADD10: print("LADD10 "); break;
      case LADD11: print("LADD11 "); break;
    }
    switch(mapram_entry & 0x600) {
      case DSIZ_32BIT: print("D32BIT "); break;
      case DSIZ_16BIT: print("D16BIT "); break;
      case DSIZ_8BIT: print("D8BIT "); break;
      case DSIZ_THREE: print("DTHREE "); break;
    }
    print("bytemux fn %x)\n", ((mapram_entry & 0x70)>>4));
    /*}}}*/
  }
  return(error || assign_integer(result_sym_p, mapram_entry, false));
}
/*}}}*/
/*{{{set_mux*/
boolean set_mux(parse_t *pars_p, char *result_sym_p)
/* sets mapram and mux contents for address range */
{
  boolean   error;
  long      base;
  long      length;
  long      def_base;
  long      def_length;
  long      def_entry;
  long      mapram_entry;
  long      read_swap;
  long      write_swap;

  /* get default values from symbol variables */
  evaluate_integer("BASEADDRESS", &def_base, 10);
  evaluate_integer("RANGE", &def_length, 10);
    
  error = cget_integer(pars_p, def_base, &base);
  if (error) 
    tag_current_line(pars_p, "invalid base address");
  else {
    def_entry = shadow_read_map_ram((void*)base);
    error = cget_integer(pars_p, def_length, &length);
    if (error)
      tag_current_line(pars_p, "invalid address range");
    else {
      error = cget_integer(pars_p, def_entry, &mapram_entry);
      if (error)
        tag_current_line(pars_p, "invalid mapram entry");
      else {
        error = cget_integer(pars_p, 0x76543210, &read_swap);
        if (error)
          tag_current_line(pars_p, "invalid readswap code");
        else {
          error = cget_integer(pars_p, 0x76543210, &write_swap);
          if (error)
            tag_current_line(pars_p, "invalid writeswap code");
          else {
            /*{{{program data into mapram and mux*/
            /* program matching mux and mapram entries for
               each 1 Mb page of memory space in wanted range*/
            int   which_swap;
            char* page;
            
            turnoff_byte_mux();
            which_swap = (int)((mapram_entry & 0x70)>>4);
            page = (char*)((unsigned)base & 0xfff00000);              
            print("Programming byte mux function %d from address #%x\n", which_swap, (int)page);
            program_byte_mux(which_swap, (int)read_swap, (int) write_swap);
            while ((long)page < (base+length)) {
              shadow_write_map_ram((void*)page, (int)mapram_entry);
              program_mux_map_ram(which_swap, (void*)page);
              page += 0x100000;
            }
            turnon_byte_mux();
            /*}}}*/
          }
        }
      }
    }
  }
  return(error || assign_integer(result_sym_p, mapram_entry, false));
}
/*}}}*/
/*{{{setup_slave*/
boolean setup_slave(parse_t *pars_p, char *result_sym_p)
/* sets decode logic for slave access to board memory */

{
  boolean   error;
  long      base;
  long      length;
  long      add_value;
  long      def_base;
  long      def_range;

  /* get default values from symbol variables */
  evaluate_integer("BASEADDRESS", &def_base, 10);
  evaluate_integer("RANGE", &def_range, 10);
    
  error = cget_integer(pars_p, def_base, &base);
  if (error) 
    tag_current_line(pars_p, "expected slave base address");
  else {
    error = (cget_integer(pars_p, def_range, &length) || (length <= 0));
    if (error)
      tag_current_line(pars_p, "invalid length");
    else {
      error = (cget_integer(pars_p, 32, &add_value) ||
               ((add_value != 32) && (add_value != 24) && (add_value !=16)));
      if (error)
        tag_current_line(pars_p, "invalid address width value");
      else {
        /* set up slave access on this region */
        prog_slave_access((void*)base, (int)length, (int)add_value, false);
      }
    }
  }
  return(error || assign_integer(result_sym_p, base, false));
}
/*}}}*/
/*{{{read_rtc*/
boolean read_rtc(parse_t *pars_p, char *result_sym_p)
/* can set a new time value into the rtc and/or display result */
{
  boolean   error;
  time_t    time_code;
  
  error = cget_integer(pars_p, -1, (long*)&time_code);
  if (error)
    tag_current_line(pars_p, "expected new time value");
  else {
    if (((long)time_code) >= 0) {
      /* program new code given into the rtc */
      print("Setting new time: %s\n", ctime(&time_code));
      rtc_settimeofday(&time_code);
    }
    else {
      /* read time as held by the clock and format to display */
      rtc_gettimeofday(&time_code);
      if((long)time_code <= 0)
        print("Odd value %d from RTC - is standby power on?\n", (long)time_code);
      else
        print("RTC clock: %s\n", ctime(&time_code));
    }
  }
  return(error || assign_integer(result_sym_p, (long)time_code, false));
}
/*}}}*/
/*{{{suspend*/
boolean suspend(parse_t *pars_p, char *result_sym_p)
/* can wait for specified number of VRTX ticks */
{
  boolean   error;
  long      time;
  
  error = cget_integer(pars_p, -1, &time);
  if (error || (time <= 0))
    tag_current_line(pars_p, "expected time tick value");
  else {
    sc_delay(time);
  }
  return(error || assign_integer(result_sym_p, time, false));
}
/*}}}*/

/*{{{console io routines for VRTX32/T environment*/
int io_write(char* buffer)
/* support for string output in a VRTX32/T supported environment 
   support padding of LF characters with CR characters for UNIX hackers */
{
   long c;
   c = (long)*buffer++;
   while(c != 0) {
     sc_putc(c);
     if(c == (long)'\n')
       sc_putc((long)'\r');
     c = (long)*buffer++;
   }

   return(0);
}

int io_read(char* prompt, char* buffer,long buflen)
/* support for a read of a null terminated input line under VRTX32/T 
   Include some minimal line editing functions for convinience */
{
  char      c;
  short     cnt;
  boolean   eol;
  long      err;

  io_write(prompt); 
  
  cnt = 0;
  eol = false;
  buffer[0] = '\0';
  c=' ';

  /* read characters one by one and perform line editing functions before
     putting them in the read buffer. Characters are echoed as accepted. */
  while((c != '\0') && !eol && (cnt < buflen)) {
    /* get character from VRTX serial buffer */
    c = (char)sc_getc();
    if (isprint(c)) {
      buffer[cnt] = c;
      cnt++;
      sc_putc((long)c, &err);
    }
    else switch(c) {
      /* control characters are recognised and dealt with. Escape
         is used to repeat the previous line of input */
      case '\r' :
      case '\n' : buffer[cnt++] = '\0';
                  sc_putc((long)'\r',&err);
                  sc_putc((long)'\n',&err);
                  eol = true;
                  strcpy(last_command, buffer);
                  break;
      case '\b' : if (cnt > 0) {
                    cnt--;
                    sc_putc((long)'\b',&err);
                    sc_putc((long)' ',&err);
                    sc_putc((long)'\b',&err);
                  }
                  break;
      case 0x1b : strcpy(buffer, last_command);
                  sc_putc((long)'\r',&err);
                  sc_putc((long)'\n',&err);
                  io_write(buffer);
                  cnt = strlen(buffer);
                  break;
    }
  }
}
/*}}}*/
/*{{{command setup*/
static boolean command_setup()
/* setup routine for IMS B016 monitor commands */
{
  register_command("DISPLAY", display_memory, 
                   "<address><range> Displays memory contents");
  register_command("SEARCH", find_memory, 
                   "<address><range><value> Searchs for memory contents");
  register_command("FILL", fill_memory, 
                   "<address><range><value> Fill memory with value");
  register_command("COPY", copy_memory, 
                   "<address><range><newaddress> Copies memory contents");
  register_command("MAPRAM", mapram, 
                   "<address> Display of mapram entry");
  register_command("MUXRAM", set_mux, 
                   "<address><range><mapcode><rswop><wswop> Set up mux & mapram entry");
  register_command("SLAVE", setup_slave, 
                   "<address><range><addwidth> Set up slave access region");
  register_command("RTC",  read_rtc, 
                   "<timeval> Sets and/or reads RTC registers");
  register_command("SUSPEND",  suspend, 
                   "<timeval> Invokes VRTX timer delay");

  /* utility default variables */
  assign_integer("MEMSIZE", 0x400000, true);
  assign_integer("BASEADDRESS", 0x80000000, false);
  assign_integer("RANGE", 0x100, false);
  assign_integer("DATAVALUE", 0, false);
  assign_integer("ADDRESSVALUE", 0x80000000, false);

  /* H/W predefines for mapram */

  assign_integer("MASTER", MASTER_PERMIT, true);
  assign_integer("D8BIT", DSIZ_8BIT, true);
  assign_integer("D16BIT", DSIZ_16BIT, true);
  assign_integer("D32BIT", DSIZ_32BIT, true);
  assign_integer("A32BIT", ASIZ_32BIT, true);
  assign_integer("A24BIT", ASIZ_24BIT, true);
  assign_integer("A16BIT", ASIZ_16BIT, true);
  assign_integer("LADD00", LADD00, true);
  assign_integer("LADD01", LADD01, true);
  assign_integer("LADD10", LADD10, true);
  assign_integer("LADD11", LADD11, true);

  /* MUX functions */
  assign_integer("FN1", 0x10, true);
  assign_integer("FN2", 0x20, true);
  assign_integer("FN3", 0x30, true);
  assign_integer("FN4", 0x40, true);
  assign_integer("FN5", 0x50, true);
  assign_integer("FN6", 0x50, true);
  assign_integer("FN7", 0x70, true);

  return(false);
}
/*}}}*/
/*{{{cli_task*/
void cli_task()
/* Print title and start interactive loop */
{
  long  err;

  while(true) {
    print("\nIMS B016/VRTX32 task level CLI v1.0\n\n");

    cli_main(command_setup, 1000, 10, "B016> ", "");
  
    sc_waitc((long)' ', &err); /* wait for wakeup character - spacebar */
  }  
}
/*}}}*/

/*{{{main*/

int main(int argc, char** argv)
{
   char     *startup_file;
   long     error;
   Process  *timer_p;
   Process  *console_p;

   /* do board level initialisation */


   basic_board_init();
   

   /* assert master role */
   init_slot_one(ARBITER_PRI,VMETIMEOUT_64US);
   init_vme_master(3, RELMODE_BCAP, false);

   /* initialise channels & start host I/O support process */
   c_input  = get_param(3);
   c_output = get_param(4); 
   rq_chr = ChanAlloc();
   rx_chr = ChanAlloc();
   tx_chr = ChanAlloc();

   console_p = ProcAlloc(console_proc, 4096,0); 
   ProcRunHigh(console_p); 

   console_output('R');
   console_output('e');
   console_output('a');
   console_output('d');
   console_output('y');
   console_output('\n');
   console_output('\n');

   /* configure VRTX and RTScope */ 
   config_vrtx();

   
   /* initialise RTL support structures */
   /* ioinit(); */

   init_break_handler();
   ProcWait(10000);
   DB_GO();


   /* do we have a startup command file? */
   if (argc > 0)
     startup_file = argv[1];
   else
     startup_file = NULL;

   /* initialise buffer for command editing support */
   strcpy(last_command,"help");

   /* start event handling code */
   start_event_handler();
   install_berr_handler(my_berr_handler);

   /* start timer isr */
   timer_p = ProcAlloc(timer_isr, 1000,0); 
   ProcRunHigh(timer_p); 

   /* initialise on board io */

   init_serial();   
   serial_setup();


   /* create the VRTX user tasks */
   sc_tcreate(cli_task,1, 2, &error);


   /* start VRTX32/T multi-tasking and never return */
   vrtx_go();

   /* if we fall out of VRTX then abort completely */
   __asm{seterr;}
}
/*}}}*/

